# -*- coding: utf-8 -*-
"""Test rules"""

import unittest

from utils import PyrseasTestCase, fix_indent, new_std_map

CREATE_TABLE_STMT = "CREATE TABLE t1 (c1 integer, c2 text)"
CREATE_STMT = "CREATE RULE r1 AS ON %s TO t1 DO %s"
COMMENT_STMT = "COMMENT ON RULE r1 ON t1 IS 'Test rule r1'"


class RuleToMapTestCase(PyrseasTestCase):
    """Test mapping of existing rules"""

    def test_map_rule_nothing(self):
        "Map a rule to do nothing"
        self.db.execute(CREATE_TABLE_STMT)
        expmap = {'r1': {'event': 'insert',
                              'actions': 'NOTHING'}}
        dbmap = self.db.execute_and_map(CREATE_STMT % ('INSERT', 'NOTHING'))
        self.assertEqual(dbmap['schema public']['table t1']['rules'],
                         expmap)

    def test_map_rule_instead(self):
        "Map rule with an INSTEAD action"
        self.db.execute(CREATE_TABLE_STMT)
        expmap = {'r1': {'event': 'update', 'instead': True,
                              'actions': 'NOTHING'}}
        dbmap = self.db.execute_and_map(CREATE_STMT % ('UPDATE',
                                                       'INSTEAD NOTHING'))
        self.assertEqual(dbmap['schema public']['table t1']['rules'],
                         expmap)

    def test_map_rule_conditional(self):
        "Map rule with a qualification"
        self.db.execute(CREATE_TABLE_STMT)
        expmap = {'r1': {
                'event': 'delete', 'condition': "(old.c1 < 1000)",
                'actions': "INSERT INTO t1 (c1, c2) VALUES (" \
                    "(old.c1 + 1000), old.c2)"}}
        dbmap = self.db.execute_and_map(
            "CREATE RULE r1 AS ON DELETE TO t1 WHERE OLD.c1 < 1000"
            "DO INSERT INTO t1 VALUES (OLD.c1 + 1000, OLD.c2)")
        self.assertEqual(dbmap['schema public']['table t1']['rules'],
                         expmap)

    def test_map_rule_multi_actions(self):
        "Map rule with multiple actions"
        self.db.execute(CREATE_TABLE_STMT)
        expmap = {'r1': {
                'event': 'update',
                'actions': "(INSERT INTO t1 (c1) VALUES ((old.c1 + 100)); "
                "INSERT INTO t1 (c1) VALUES ((old.c1 + 200)); )"}}
        dbmap = self.db.execute_and_map(CREATE_STMT % (
                'UPDATE', "(INSERT INTO t1 (c1) VALUES (old.c1 + 100); "
                "INSERT INTO t1 (c1) VALUES (old.c1 + 200))"))
        self.assertEqual(dbmap['schema public']['table t1']['rules'],
                         expmap)

    def test_map_rule_comment(self):
        "Map a rule comment"
        self.db.execute(CREATE_TABLE_STMT)
        self.db.execute(CREATE_STMT % ('INSERT', 'NOTHING'))
        dbmap = self.db.execute_and_map(COMMENT_STMT)
        self.assertEqual(dbmap['schema public']['table t1']['rules']
                         ['r1']['description'], 'Test rule r1')


class RuleToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input rules"""

    def test_create_rule_nothing(self):
        "Create a rule"
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'rules': {'r1': {'event': 'insert',
                                     'actions': 'NOTHING'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), CREATE_TABLE_STMT)
        self.assertEqual(fix_indent(dbsql[1]), CREATE_STMT % (
                'INSERT', 'NOTHING'))

    def test_create_rule_instead(self):
        "Create a rule with an INSTEAD action"
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'rules': {'r1': {'event': 'update', 'instead': True,
                                     'actions': 'NOTHING'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), CREATE_TABLE_STMT)
        self.assertEqual(fix_indent(dbsql[1]),
                         "CREATE RULE r1 AS ON UPDATE TO t1 "
                         "DO INSTEAD NOTHING")

    def test_create_rule_conditional(self):
        "Create a rule with qualification"
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'rules': {'r1': {'event': 'delete',
                                     'condition': "old.c1 < 1000",
                                     'actions': "INSERT INTO t1 VALUES (" \
                                         "old.c1 + 1000, old.c2)"}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), CREATE_TABLE_STMT)
        self.assertEqual(fix_indent(dbsql[1]), "CREATE RULE r1 AS ON DELETE "
                         "TO t1 WHERE old.c1 < 1000 "
                         "DO INSERT INTO t1 VALUES (old.c1 + 1000, old.c2)")

    def test_create_rule_multi_actions(self):
        "Create a rule with multiple actions"
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'rules': {'r1': {
                            'event': 'update', 'actions':
                                "(INSERT INTO t1 VALUES (old.c1 + 100); " \
                                "INSERT INTO t1 VALUES (old.c1 + 200));)"}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), CREATE_TABLE_STMT)
        self.assertEqual(fix_indent(dbsql[1]), "CREATE RULE r1 AS ON UPDATE "
                         "TO t1 DO (INSERT INTO t1 VALUES (old.c1 + 100); "
                         "INSERT INTO t1 VALUES (old.c1 + 200));)")

    def test_create_rule_in_schema(self):
        "Create a rule within a non-public schema"
        self.db.execute("DROP SCHEMA IF EXISTS s1 CASCADE")
        self.db.execute_commit("CREATE SCHEMA s1")
        inmap = new_std_map()
        inmap.update({'schema s1': {
                    'table t1': {
                        'columns': [{'c1': {'type': 'integer'}},
                                    {'c2': {'type': 'text'}}],
                    'rules': {'r1': {'event': 'insert',
                                     'actions': 'NOTHING'}}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[1]), "CREATE RULE r1 "
                         "AS ON INSERT TO s1.t1 DO NOTHING")
        self.db.execute_commit("DROP SCHEMA s1 CASCADE")

    def test_drop_rule(self):
        "Drop an existing rule"
        self.db.execute(CREATE_TABLE_STMT)
        self.db.execute_commit(CREATE_STMT % ('INSERT', 'NOTHING'))
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}]}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["DROP RULE r1 ON t1"])

    def test_drop_rule_table(self):
        "Drop an existing rule and the related table"
        self.db.execute(CREATE_TABLE_STMT)
        self.db.execute_commit(CREATE_STMT % ('INSERT', 'NOTHING'))
        dbsql = self.db.process_map(new_std_map())
        self.assertEqual(dbsql[0], "DROP RULE r1 ON t1")
        self.assertEqual(dbsql[1], "DROP TABLE t1")

    def test_rule_with_comment(self):
        "Create a rule with a comment"
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'rules': {'r1': {'event': 'insert',
                                     'description': 'Test rule r1',
                                     'actions': 'NOTHING'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql[2], COMMENT_STMT)

    def test_comment_on_rule(self):
        "Create a comment on an existing rule"
        self.db.execute(CREATE_TABLE_STMT)
        self.db.execute_commit(CREATE_STMT % ('INSERT', 'NOTHING'))
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'rules': {'r1': {'event': 'insert',
                                     'description': 'Test rule r1',
                                     'actions': 'NOTHING'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, [COMMENT_STMT])

    def test_drop_rule_comment(self):
        "Drop the comment on an existing rule"
        self.db.execute(CREATE_TABLE_STMT)
        self.db.execute(CREATE_STMT % ('INSERT', 'NOTHING'))
        self.db.execute_commit(COMMENT_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'rules': {'r1': {'event': 'insert',
                                     'actions': 'NOTHING'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql,
                         ["COMMENT ON RULE r1 ON t1 IS NULL"])

    def test_change_rule_comment(self):
        "Change existing comment on a rule"
        self.db.execute(CREATE_TABLE_STMT)
        self.db.execute(CREATE_STMT % ('INSERT', 'NOTHING'))
        self.db.execute_commit(COMMENT_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'rules': {'r1': {'event': 'insert',
                                     'description': 'Changed rule r1',
                                     'actions': 'NOTHING'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, [
                "COMMENT ON RULE r1 ON t1 IS 'Changed rule r1'"])


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(RuleToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            RuleToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
    pyrseas.language
    ~~~~~~~~~~~~~~~~

    This defines two classes, Language and LanguageDict, derived from
    DbObject and DbObjectDict, respectively.
"""
from pyrseas.dbobject import DbObjectDict, DbObject, quote_id
from pyrseas.dbobject.function import Function


class Language(DbObject):
    """A procedural language definition"""

    keylist = ['name']
    objtype = "LANGUAGE"

    def to_map(self):
        """Convert language to a YAML-suitable format

        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        if 'functions' in dct:
            del dct['functions']
        key = self.extern_key()
        language = {key: dct}
        if hasattr(self, 'description'):
            language[key].update(description=self.description)
        return language

    def create(self):
        """Return SQL statements to CREATE the language

        :return: SQL statements
        """
        stmts = ["CREATE LANGUAGE %s" % quote_id(self.name)]
        if hasattr(self, 'description'):
            stmts.append(self.comment())
        return stmts

    def diff_map(self, inlanguage):
        """Generate SQL to transform an existing language

        :param inlanguage: a YAML map defining the new language
        :return: list of SQL statements

        Compares the language to an input language and generates SQL
        statements to transform it into the one represented by the
        input.
        """
        stmts = []
        if hasattr(self, 'description'):
            if hasattr(inlanguage, 'description'):
                if self.description != inlanguage.description:
                    self.description = inlanguage.description
                    stmts.append(self.comment())
            else:
                del self.description
                stmts.append(self.comment())
        else:
            if hasattr(inlanguage, 'description'):
                self.description = inlanguage.description
                stmts.append(self.comment())
        return stmts


class LanguageDict(DbObjectDict):
    "The collection of procedural languages in a database."

    cls = Language
    query = \
        """SELECT lanname AS name, lanpltrusted AS trusted, description
           FROM pg_language l
                LEFT JOIN pg_description d
                     ON (l.oid = d.objoid AND d.objsubid = 0)
           WHERE lanispl
           ORDER BY lanname"""

    def from_map(self, inmap):
        """Initialize the dictionary of languages by examining the input map

        :param inmap: the input YAML map defining the languages
        """
        for lng in inmap.keys():
            spc = lng.find(' ')
            if spc == -1:
                raise KeyError("Unrecognized object type: %s" % lng)
            key = lng[spc + 1:]
            language = self[key] = Language(name=key)
            inlanguage = inmap[lng]
            if inlanguage:
                if 'oldname' in inlanguage:
                    language.oldname = inlanguage['oldname']
                    del inlanguage['oldname']
                if 'description' in inlanguage:
                    language.description = inlanguage['description']

    def link_refs(self, dbfunctions):
        """Connect functions to their respective languages

        :param dbfunctions: dictionary of functions

        Fills in the `functions` dictionary for each language by
        traversing the `dbfunctions` dictionary, which is keyed by
        schema and function name.
        """
        for (sch, fnc, arg) in dbfunctions.keys():
            func = dbfunctions[(sch, fnc, arg)]
            if func.language == 'sql' or func.language == 'internal':
                continue
            assert self[(func.language)]
            language = self[(func.language)]
            if isinstance(func, Function):
                if not hasattr(language, 'functions'):
                    language.functions = {}
                language.functions.update({fnc: func})

    def to_map(self):
        """Convert the language dictionary to a regular dictionary

        :return: dictionary

        Invokes the `to_map` method of each language to construct a
        dictionary of languages.
        """
        languages = {}
        for lng in self.keys():
            languages.update(self[lng].to_map())
        return languages

    def diff_map(self, inlanguages, dbversion):
        """Generate SQL to transform existing languages

        :param input_map: a YAML map defining the new languages
        :param dbversion: DBMS version number
        :return: list of SQL statements

        Compares the existing language definitions, as fetched from the
        catalogs, to the input map and generates SQL statements to
        transform the languages accordingly.
        """
        stmts = []
        # check input languages
        for lng in inlanguages.keys():
            inlng = inlanguages[lng]
            # does it exist in the database?
            if lng in self:
                stmts.append(self[lng].diff_map(inlng))
            else:
                # check for possible RENAME
                if hasattr(inlng, 'oldname'):
                    oldname = inlng.oldname
                    try:
                        stmts.append(self[oldname].rename(inlng.name))
                        del self[oldname]
                    except KeyError, exc:
                        exc.args = ("Previous name '%s' for language '%s' "
                                   "not found" % (oldname, inlng.name), )
                        raise
                else:
                    # create new language
                    stmts.append(inlng.create())
        # check database languages
        for lng in self.keys():
            # if missing, drop it
            if lng not in inlanguages:
                # special case: plpgsql is installed in 9.0
                if dbversion >= 90000 and self[lng].name == 'plpgsql':
                    continue
                self[lng].dropped = True
        return stmts

    def _drop(self):
        """Actually drop the languages

        :return: SQL statements
        """
        stmts = []
        for lng in self.keys():
            if hasattr(self[lng], 'dropped'):
                stmts.append(self[lng].drop())
        return stmts



# -*- coding: utf-8 -*-
"""
    pyrseas.schema
    ~~~~~~~~~~~~~~

    This defines two classes, Schema and SchemaDict, derived from
    DbObject and DbObjectDict, respectively.
"""
from pyrseas.dbobject import DbObjectDict, DbObject
from pyrseas.dbobject import quote_id, split_schema_table
from dbtype import Domain, Enum
from table import Table, Sequence, View


class Schema(DbObject):
    """A database schema definition, i.e., a named collection of tables,
    views, triggers and other schema objects."""

    keylist = ['name']
    objtype = 'SCHEMA'

    def to_map(self, dbschemas):
        """Convert tables, etc., dictionaries to a YAML-suitable format

        :param dbschemas: dictionary of schemas
        :return: dictionary
        """
        key = self.extern_key()
        schema = {key: {}}
        if hasattr(self, 'domains'):
            doms = {}
            for dom in self.domains.keys():
                doms.update(self.domains[dom].to_map())
            schema[key].update(doms)
        if hasattr(self, 'types'):
            typs = {}
            for typ in self.types.keys():
                typs.update(self.types[typ].to_map())
            schema[key].update(typs)
        if hasattr(self, 'sequences'):
            seqs = {}
            for seq in self.sequences.keys():
                seqs.update(self.sequences[seq].to_map())
            schema[key].update(seqs)
        if hasattr(self, 'tables'):
            tbls = {}
            for tbl in self.tables.keys():
                tbls.update(self.tables[tbl].to_map(dbschemas))
            schema[key].update(tbls)
        if hasattr(self, 'views'):
            views = {}
            for view in self.views.keys():
                views.update(self.views[view].to_map())
            schema[key].update(views)
        if hasattr(self, 'functions'):
            functions = {}
            for func in self.functions.keys():
                functions.update(self.functions[func].to_map())
            schema[key].update(functions)
        if hasattr(self, 'description'):
            schema[key].update(description=self.description)
        return schema

    def create(self):
        """Return SQL statements to CREATE the schema

        :return: SQL statements
        """
        stmts = ["CREATE SCHEMA %s" % quote_id(self.name)]
        if hasattr(self, 'description'):
            stmts.append(self.comment())
        return stmts

    def diff_map(self, inschema):
        """Generate SQL to transform an existing schema

        :param inschema: a YAML map defining the new schema
        :return: list of SQL statements

        Compares the schema to an input schema and generates SQL
        statements to transform it into the one represented by the
        input.
        """
        stmts = []
        if hasattr(self, 'description'):
            if hasattr(inschema, 'description'):
                if self.description != inschema.description:
                    self.description = inschema.description
                    stmts.append(self.comment())
            else:
                del self.description
                stmts.append(self.comment())
        else:
            if hasattr(inschema, 'description'):
                self.description = inschema.description
                stmts.append(self.comment())
        return stmts


class SchemaDict(DbObjectDict):
    "The collection of schemas in a database.  Minimally, the 'public' schema."

    cls = Schema
    query = \
        """SELECT nspname AS name, description
           FROM pg_namespace n JOIN pg_roles ON (nspowner = pg_roles.oid)
                LEFT JOIN pg_description d
                     ON (n.oid = d.objoid AND d.objsubid = 0)
           WHERE nspname = 'public' OR rolname <> 'postgres'
           ORDER BY nspname"""

    def from_map(self, inmap, newdb):
        """Initialize the dictionary of schemas by converting the input map

        :param inmap: the input YAML map defining the schemas
        :param newdb: collection of dictionaries defining the database

        Starts the recursive analysis of the input map and
        construction of the internal collection of dictionaries
        describing the database objects.
        """
        for sch in inmap.keys():
            spc = sch.find(' ')
            if spc == -1:
                raise KeyError("Unrecognized object type: %s" % sch)
            key = sch[spc + 1:]
            schema = self[key] = Schema(name=key)
            inschema = inmap[sch]
            intypes = {}
            intables = {}
            infuncs = {}
            for key in inschema.keys():
                if key.startswith('domain '):
                    intypes.update({key: inschema[key]})
                elif key.startswith('type '):
                    intypes.update({key: inschema[key]})
                elif key.startswith('table ') or key.startswith('view ') \
                        or key.startswith('sequence '):
                    intables.update({key: inschema[key]})
                elif key.startswith('function ') \
                        or key.startswith('aggregate '):
                    infuncs.update({key: inschema[key]})
                elif key == 'oldname':
                    schema.oldname = inschema[key]
                    del inschema['oldname']
                elif key == 'description':
                    schema.description = inschema[key]
                else:
                    raise KeyError("Expected typed object, found '%s'" % key)
            newdb.types.from_map(schema, intypes, newdb)
            newdb.tables.from_map(schema, intables, newdb)
            newdb.functions.from_map(schema, infuncs)

    def link_refs(self, dbtypes, dbtables, dbfunctions):
        """Connect types, tables and functions to their respective schemas

        :param dbtypes: dictionary of types and domains
        :param dbtables: dictionary of tables, sequences and views
        :param dbfunctions: dictionary of functions

        Fills in the `domains` dictionary for each schema by
        traversing the `dbtypes` dictionary.  Fills in the `tables`,
        `sequences`, `views` dictionaries for each schema by
        traversing the `dbtables` dictionary. Fills in the `functions`
        dictionary by traversing the `dbfunctions` dictionary.
        """
        for (sch, typ) in dbtypes.keys():
            dbtype = dbtypes[(sch, typ)]
            assert self[sch]
            schema = self[sch]
            if isinstance(dbtype, Domain):
                if not hasattr(schema, 'domains'):
                    schema.domains = {}
                schema.domains.update({typ: dbtypes[(sch, typ)]})
            elif isinstance(dbtype, Enum):
                if not hasattr(schema, 'types'):
                    schema.types = {}
                schema.types.update({typ: dbtypes[(sch, typ)]})
        for (sch, tbl) in dbtables.keys():
            table = dbtables[(sch, tbl)]
            assert self[sch]
            schema = self[sch]
            if isinstance(table, Table):
                if not hasattr(schema, 'tables'):
                    schema.tables = {}
                schema.tables.update({tbl: table})
            elif isinstance(table, Sequence):
                if not hasattr(schema, 'sequences'):
                    schema.sequences = {}
                schema.sequences.update({tbl: table})
            elif isinstance(table, View):
                if not hasattr(schema, 'views'):
                    schema.views = {}
                schema.views.update({tbl: table})
        for (sch, fnc, arg) in dbfunctions.keys():
            func = dbfunctions[(sch, fnc, arg)]
            assert self[sch]
            schema = self[sch]
            if not hasattr(schema, 'functions'):
                schema.functions = {}
            schema.functions.update({(fnc, arg): func})
            if hasattr(func, 'returns'):
                rettype = func.returns
                if rettype.upper().startswith("SETOF "):
                    rettype = rettype[6:]
                (retsch, rettyp) = split_schema_table(rettype, sch)
                if (retsch, rettyp) in dbtables.keys():
                    deptbl = dbtables[(retsch, rettyp)]
                    if not hasattr(func, 'dependent_table'):
                        func.dependent_table = deptbl
                    if not hasattr(deptbl, 'dependent_funcs'):
                        deptbl.dependent_funcs = []
                    deptbl.dependent_funcs.append(func)

    def to_map(self):
        """Convert the schema dictionary to a regular dictionary

        :return: dictionary

        Invokes the `to_map` method of each schema to construct a
        dictionary of schemas.
        """
        schemas = {}
        for sch in self.keys():
            schemas.update(self[sch].to_map(self))
        return schemas

    def diff_map(self, inschemas):
        """Generate SQL to transform existing schemas

        :param input_map: a YAML map defining the new schemas
        :return: list of SQL statements

        Compares the existing schema definitions, as fetched from the
        catalogs, to the input map and generates SQL statements to
        transform the schemas accordingly.
        """
        stmts = []
        # check input schemas
        for sch in inschemas.keys():
            insch = inschemas[sch]
            # does it exist in the database?
            if sch in self:
                stmts.append(self[sch].diff_map(insch))
            else:
                # check for possible RENAME
                if hasattr(insch, 'oldname'):
                    oldname = insch.oldname
                    try:
                        stmts.append(self[oldname].rename(insch.name))
                        del self[oldname]
                    except KeyError, exc:
                        exc.args = ("Previous name '%s' for schema '%s' "
                                   "not found" % (oldname, insch.name), )
                        raise
                else:
                    # create new schema
                    stmts.append(insch.create())
        # check database schemas
        for sch in self.keys():
            # if missing and not 'public', drop it
            if sch != 'public' and sch not in inschemas:
                self[sch].dropped = True
        return stmts

    def _drop(self):
        """Actually drop the schemas

        :return: SQL statements
        """
        stmts = []
        for sch in self.keys():
            if sch != 'public' and hasattr(self[sch], 'dropped'):
                stmts.append(self[sch].drop())
        return stmts



# -*- coding: utf-8 -*-
"""Pyrseas unit tests"""

import unittest

import test_language
import test_schema
import test_type
import test_domain
import test_sequence
import test_table
import test_constraint
import test_index
import test_view
import test_function
import test_trigger
import test_rule


def suite():
    tests = unittest.TestSuite()
    tests.addTest(test_language.suite())
    tests.addTest(test_schema.suite())
    tests.addTest(test_type.suite())
    tests.addTest(test_domain.suite())
    tests.addTest(test_sequence.suite())
    tests.addTest(test_table.suite())
    tests.addTest(test_constraint.suite())
    tests.addTest(test_index.suite())
    tests.addTest(test_view.suite())
    tests.addTest(test_function.suite())
    tests.addTest(test_trigger.suite())
    tests.addTest(test_rule.suite())
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""Test sequences"""

import unittest

from utils import PyrseasTestCase, fix_indent, new_std_map

DROP_STMT = "DROP SEQUENCE IF EXISTS seq1"


class SequenceToMapTestCase(PyrseasTestCase):
    """Test mapping of created sequences"""

    def test_map_sequence(self):
        "Map a created sequence"
        ddlstmt = "CREATE SEQUENCE seq1"
        expmap = {'start_value': 1, 'increment_by': 1, 'max_value': None,
                  'min_value': None, 'cache_value': 1}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['sequence seq1'],
                         expmap)


class SequenceToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input sequences"""

    def test_create_sequence(self):
        "Create a sequence"
        self.db.execute_commit(DROP_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'sequence seq1': {
                    'start_value': 1, 'increment_by': 1, 'max_value': None,
                    'min_value': None, 'cache_value': 1}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "CREATE SEQUENCE seq1 START WITH 1 "
                         "INCREMENT BY 1 NO MAXVALUE NO MINVALUE CACHE 1")

    def test_create_sequence_in_schema(self):
        "Create a sequence within a non-public schema"
        self.db.execute("CREATE SCHEMA s1")
        self.db.execute_commit(DROP_STMT)
        inmap = new_std_map()
        inmap.update({'schema s1': {'sequence seq1': {
                    'start_value': 1, 'increment_by': 1, 'max_value': None,
                    'min_value': None, 'cache_value': 1}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "CREATE SEQUENCE s1.seq1 START WITH 1 "
                         "INCREMENT BY 1 NO MAXVALUE NO MINVALUE CACHE 1")
        self.db.execute_commit("DROP SCHEMA s1 CASCADE")

    def test_bad_sequence_map(self):
        "Error creating a sequence with a bad map"
        inmap = new_std_map()
        inmap['schema public'].update({'seq1': {
                    'start_value': 1, 'increment_by': 1, 'max_value': None,
                    'min_value': None, 'cache_value': 1}})
        self.assertRaises(KeyError, self.db.process_map, inmap)

    def test_drop_sequence(self):
        "Drop an existing sequence"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE SEQUENCE seq1")
        dbsql = self.db.process_map(new_std_map())
        self.assertEqual(dbsql, ["DROP SEQUENCE seq1"])

    def test_rename_sequence(self):
        "Rename an existing sequence"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE SEQUENCE seq1")
        inmap = new_std_map()
        inmap['schema public'].update({'sequence seq2': {
                    'oldname': 'seq1',
                    'start_value': 1, 'increment_by': 1, 'max_value': None,
                    'min_value': None, 'cache_value': 1}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["ALTER SEQUENCE seq1 RENAME TO seq2"])

    def test_bad_rename_sequence(self):
        "Error renaming a non-existing sequence"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE SEQUENCE seq1")
        inmap = new_std_map()
        inmap['schema public'].update({'sequence seq2': {
                    'oldname': 'seq3',
                    'start_value': 1, 'increment_by': 1, 'max_value': None,
                    'min_value': None, 'cache_value': 1}})
        self.assertRaises(KeyError, self.db.process_map, inmap)

    def test_change_sequence(self):
        "Change sequence attributes"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE SEQUENCE seq1")
        inmap = new_std_map()
        inmap['schema public'].update({'sequence seq1': {
                    'start_value': 5, 'increment_by': 10, 'max_value': None,
                    'min_value': None, 'cache_value': 30}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "ALTER SEQUENCE seq1 START WITH 5 INCREMENT BY 10 "
                         "CACHE 30")


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(SequenceToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            SequenceToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
    pyrseas.trigger
    ~~~~~~~~~~~~~~~

    This module defines two classes: Trigger derived from
    DbSchemaObject, and TriggerDict derived from DbObjectDict.
"""
from pyrseas.dbobject import DbObjectDict, DbSchemaObject, quote_id


EXEC_PROC = 'EXECUTE PROCEDURE '
EVENT_TYPES = ['INSERT', 'UPDATE', 'DELETE', 'TRUNCATE']


class Trigger(DbSchemaObject):
    """A procedural language trigger"""

    keylist = ['schema', 'table', 'name']
    objtype = "TRIGGER"

    def identifier(self):
        """Returns a full identifier for the trigger

        :return: string
        """
        return "%s ON %s" % (quote_id(self.name), self._table.qualname())

    def to_map(self):
        """Convert a trigger to a YAML-suitable format

        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        del dct['_table']
        if hasattr(self, 'columns'):
            dct['columns'] = [self._table.column_names()[int(k) - 1]
                              for k in self.columns.split()]
        return {self.name: dct}

    def create(self):
        """Return SQL statements to CREATE the trigger

        :return: SQL statements
        """
        stmts = []
        evts = " OR ".join(self.events).upper()
        if hasattr(self, 'columns') and 'update' in self.events:
            evts = evts.replace("UPDATE", "UPDATE OF %s" % (
                    ", ".join(self.columns)))
        cond = ''
        if hasattr(self, 'condition'):
            cond = "\n    WHEN (%s)" % self.condition
        stmts.append("CREATE TRIGGER %s\n    %s %s ON %s\n    FOR EACH %s"
                     "%s\n    EXECUTE PROCEDURE %s" % (
                quote_id(self.name), self.timing.upper(), evts,
                self._table.qualname(), self.level.upper(), cond,
                self.procedure))
        if hasattr(self, 'description'):
            stmts.append(self.comment())
        return stmts

    def diff_map(self, intrigger):
        """Generate SQL to transform an existing trigger

        :param intrigger: a YAML map defining the new trigger
        :return: list of SQL statements

        Compares the trigger to an input trigger and generates SQL
        statements to transform it into the one represented by the
        input. For now, only the description can be changed.
        """
        stmts = []
        stmts.append(self.diff_description(intrigger))
        return stmts


QUERY_PRE90 = \
        """SELECT nspname AS schema, relname AS table,
                  tgname AS name, pg_get_triggerdef(t.oid) AS definition,
                  NULL AS columns, description
           FROM pg_trigger t
                JOIN pg_class c ON (t.tgrelid = c.oid)
                JOIN pg_namespace n ON (c.relnamespace = n.oid)
                JOIN pg_roles ON (n.nspowner = pg_roles.oid)
                LEFT JOIN pg_description d
                     ON (t.oid = d.objoid AND d.objsubid = 0)
           WHERE NOT tgisconstraint
             AND (nspname != 'pg_catalog' AND nspname != 'information_schema')
           ORDER BY 1, 2, 3"""


class TriggerDict(DbObjectDict):
    "The collection of triggers in a database"

    cls = Trigger
    query = \
        """SELECT nspname AS schema, relname AS table,
                  tgname AS name, pg_get_triggerdef(t.oid) AS definition,
                  tgattr AS columns, description
           FROM pg_trigger t
                JOIN pg_class c ON (t.tgrelid = c.oid)
                JOIN pg_namespace n ON (c.relnamespace = n.oid)
                JOIN pg_roles ON (n.nspowner = pg_roles.oid)
                LEFT JOIN pg_description d
                     ON (t.oid = d.objoid AND d.objsubid = 0)
           WHERE NOT tgisinternal
             AND (nspname != 'pg_catalog' AND nspname != 'information_schema')
           ORDER BY 1, 2, 3"""

    def _from_catalog(self):
        """Initialize the dictionary of triggers by querying the catalogs"""
        if self.dbconn.version < 90000:
            self.query = QUERY_PRE90
        for trig in self.fetch():
            if 'BEFORE ' in trig.definition:
                trig.timing = 'before'
                evtstart = trig.definition.index('BEFORE ') + 7
            else:
                trig.timing = 'after'
                evtstart = trig.definition.index('AFTER ') + 6
            evtend = trig.definition.index(' ON ', evtstart)
            events = trig.definition[evtstart:evtend]
            trig.events = []
            for evt in EVENT_TYPES:
                if evt in events:
                    trig.events.append(evt.lower())
            trig.level = ('FOR EACH ROW' in trig.definition and 'row' or
                          'statement')
            if 'WHEN (' in trig.definition:
                trig.condition = trig.definition[
                    trig.definition.index('WHEN (') + 6:
                        trig.definition.index(') EXECUTE PROCEDURE')]
            trig.procedure = trig.definition[trig.definition.index(EXEC_PROC)
                                             + len(EXEC_PROC):]
            del trig.definition
            self[trig.key()] = trig

    def from_map(self, table, intriggers):
        """Initalize the dictionary of triggers by converting the input map

        :param table: table owning the triggers
        :param intriggers: YAML map defining the triggers
        """
        for trg in intriggers.keys():
            intrig = intriggers[trg]
            if not intrig:
                raise ValueError("Trigger '%s' has no specification" % trg)
            self[(table.schema, table.name, trg)] = trig = Trigger(
                schema=table.schema, table=table.name, name=trg)
            for attr, val in intrig.items():
                setattr(trig, attr, val)
            if not hasattr(trig, 'level'):
                trig.level = 'statement'
            if 'oldname' in intrig:
                trig.oldname = intrig['oldname']
            if 'description' in intrig:
                trig.description = intrig['description']

    def diff_map(self, intriggers):
        """Generate SQL to transform existing triggers

        :param intriggers: a YAML map defining the new triggers
        :return: list of SQL statements

        Compares the existing trigger definitions, as fetched from
        the catalogs, to the input map and generates SQL statements to
        transform the triggers accordingly.
        """
        stmts = []
        # check input triggers
        for (sch, tbl, trg) in intriggers.keys():
            intrig = intriggers[(sch, tbl, trg)]
            # does it exist in the database?
            if (sch, tbl, trg) not in self:
                if not hasattr(intrig, 'oldname'):
                    # create new trigger
                    stmts.append(intrig.create())
                else:
                    stmts.append(self[(sch, tbl, trg)].rename(intrig))
            else:
                # check trigger objects
                stmts.append(self[(sch, tbl, trg)].diff_map(intrig))

        # check existing triggers
        for (sch, tbl, trg) in self.keys():
            trig = self[(sch, tbl, trg)]
            # if missing, drop them
            if (sch, tbl, trg) not in intriggers:
                    stmts.append(trig.drop())

        return stmts



# -*- coding: utf-8 -*-
"""
    pyrseas.dbobject
    ~~~~~~~~~~~~~~~~

    This defines two low level classes and an intermediate class.
    Most Pyrseas classes are derived from either DbObject or
    DbObjectDict.
"""
import string


VALID_FIRST_CHARS = string.lowercase + '_'
VALID_CHARS = string.lowercase + string.digits + '_$'


def quote_id(name):
    """Quotes an identifier if necessary.

    :param name: string to be quoted

    :return: possibly quoted string
    """
    # TODO: keywords
    regular_id = True
    if not name[0] in VALID_FIRST_CHARS:
        regular_id = False
    else:
        for ltr in name[1:]:
            if ltr not in VALID_CHARS:
                regular_id = False
                break

    return regular_id and name or '"%s"' % name


def split_schema_table(tbl, sch=None):
    """Return a (schema, table) tuple given a possibly schema-qualified name

    :param tbl: table name or schema.table
    :return: tuple
    """
    qualsch = sch
    if sch == None:
        qualsch = 'public'
    if '.' in tbl:
        (qualsch, tbl) = tbl.split('.')
    if tbl[:1] == '"' and tbl[-1:] == '"':
        tbl = tbl[1:-1]
    if sch != qualsch:
        sch = qualsch
    return (sch, tbl)


class DbObject(object):
    "A single object in a database catalog, e.g., a schema, a table, a column"

    keylist = []
    objtype = ''

    def __init__(self, **attrs):
        """Initialize the catalog object from a dictionary of attributes

        :param attrs: the dictionary of attributes

        Non-key attributes without a value are discarded.
        """
        for key, val in attrs.items():
            if val or key in self.keylist:
                setattr(self, key, val)

    def extern_key(self):
        """Return the key to be used in external maps for this object

        :return: string
        """
        return '%s %s' % (self.objtype.lower(), self.name)

    def key(self):
        """Return a tuple that identifies the database object

        :return: a single value or a tuple
        """
        lst = [getattr(self, k) for k in self.keylist]
        return len(lst) == 1 and lst[0] or tuple(lst)

    def identifier(self):
        """Returns a full identifier for the database object

        :return: string
        """
        return quote_id(self.name)

    def comment(self):
        """Return SQL statement to create COMMENT on object

        :return: SQL statement
        """
        if hasattr(self, 'description'):
            descr = "'%s'" % self.description
        else:
            descr = 'NULL'
        return "COMMENT ON %s %s IS %s" % (
            self.objtype, self.identifier(), descr)

    def drop(self):
        """Return SQL statement to DROP the object

        :return: SQL statement
        """
        return "DROP %s %s" % (self.objtype, self.identifier())

    def rename(self, newname):
        """Return SQL statement to RENAME the object

        :param newname: the new name for the object
        :return: SQL statement
        """
        return "ALTER %s %s RENAME TO %s" % (self.objtype, self.name, newname)

    def diff_description(self, inobj):
        """Generate SQL statements to add or change COMMENTs

        :param inobj: a YAML map defining the input object
        :return: list of SQL statements
        """
        stmts = []
        if hasattr(self, 'description'):
            if hasattr(inobj, 'description'):
                if self.description != inobj.description:
                    self.description = inobj.description
                    stmts.append(self.comment())
            else:
                del self.description
                stmts.append(self.comment())
        else:
            if hasattr(inobj, 'description'):
                self.description = inobj.description
                stmts.append(self.comment())
        return stmts


class DbSchemaObject(DbObject):
    "A database object that is owned by a certain schema"

    def identifier(self):
        """Return a full identifier for a schema object

        :return: string
        """
        return self.qualname()

    def qualname(self):
        """Return the schema-qualified name of the object

        :return: string

        No qualification is used if the schema is 'public'.
        """
        return self.schema == 'public' and quote_id(self.name) \
            or "%s.%s" % (quote_id(self.schema), quote_id(self.name))

    def unqualify(self):
        """Adjust the schema and table name if the latter is qualified"""
        if hasattr(self, 'table') and '.' in self.table:
            (sch, self.table) = split_schema_table(self.table, self.schema)

    def drop(self):
        """Return a SQL DROP statement for the schema object

        :return: SQL statement
        """
        if not hasattr(self, 'dropped') or not self.dropped:
            self.dropped = True
            return "DROP %s %s" % (self.objtype, self.identifier())
        return []

    def rename(self, newname):
        """Return a SQL ALTER statement to RENAME the schema object

        :param newname: the new name of the object
        :return: SQL statement
        """
        return "ALTER %s %s RENAME TO %s" % (self.objtype, self.qualname(),
                                             newname)

    def set_search_path(self):
        """Return a SQL SET search_path if not in the 'public' schema"""
        stmt = ''
        if self.schema != 'public':
            stmt = "SET search_path TO %s, pg_catalog" % quote_id(self.schema)
        return stmt


class DbObjectDict(dict):
    """A dictionary of database objects, all of the same type"""

    cls = DbObject
    query = ''

    def __init__(self, dbconn=None):
        """Initialize the dictionary

        :param dbconn: a DbConnection object

        If dbconn is not None, the _from_catalog method is called to
        initialize the dictionary from the catalogs.
        """
        dict.__init__(self)
        self.dbconn = dbconn
        if dbconn:
            self._from_catalog()

    def _from_catalog(self):
        """Initialize the dictionary by querying the catalogs

        This is may be overriden by derived classes as needed.
        """
        for obj in self.fetch():
            self[obj.key()] = obj

    def fetch(self):
        """Fetch all objects from the catalogs using the class query

        :return: list of self.cls objects
        """
        if not self.dbconn.conn:
            self.dbconn.connect()
        data = self.dbconn.fetchall(self.query)
        return [self.cls(**dict(row)) for row in data]



# -*- coding: utf-8 -*-
"""
    pyrseas.index
    ~~~~~~~~~~~~~

    This defines two classes, Index and IndexDict, derived
    from DbSchemaObject and DbObjectDict, respectively.
"""
from pyrseas.dbobject import DbObjectDict, DbSchemaObject, quote_id


class Index(DbSchemaObject):
    """A physical index definition, other than a primary key or unique
    constraint index.
    """

    keylist = ['schema', 'table', 'name']
    objtype = "INDEX"

    def key_columns(self):
        """Return comma-separated list of key column names

        :return: string
        """
        return ", ".join(self.keycols)

    def to_map(self, dbcols):
        """Convert an index definition to a YAML-suitable format

        :param dbcols: dictionary of dbobject columns
        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        if hasattr(self, 'keycols'):
            dct['columns'] = [dbcols[int(k) - 1] for k in self.keycols.split()]
            del dct['keycols']
        return {self.name: dct}

    def create(self):
        """Return a SQL statement to CREATE the index

        :return: SQL statements
        """
        stmts = []
        pth = self.set_search_path()
        if pth:
            stmts.append(pth)
        unq = hasattr(self, 'unique') and self.unique
        acc = hasattr(self, 'access_method') \
            and 'USING %s ' % self.access_method or ''
        stmts.append("CREATE %sINDEX %s ON %s %s(%s)" % (
            unq and 'UNIQUE ' or '', quote_id(self.name), quote_id(self.table),
            acc, hasattr(self, 'keycols') and self.key_columns() or
            self.expression))
        return stmts

    def diff_map(self, inindex):
        """Generate SQL to transform an existing index

        :param inindex: a YAML map defining the new index
        :return: list of SQL statements

        Compares the index to an input index and generates SQL
        statements to transform it into the one represented by the
        input.
        """
        stmts = []
        if not hasattr(self, 'unique'):
            self.unique = False
        if self.access_method != inindex.access_method \
                or self.unique != inindex.unique:
            stmts.append("DROP INDEX %s" % self.qualname())
            self.access_method = inindex.access_method
            self.unique = inindex.unique
            stmts.append(self.create())
        # TODO: need to deal with changes in keycols
        return stmts


class IndexDict(DbObjectDict):
    "The collection of indexes on tables in a database"

    cls = Index
    query = \
        """SELECT nspname AS schema, indrelid::regclass AS table,
                  c.relname AS name, amname AS access_method,
                  indisunique AS unique, indkey AS keycols,
                  pg_get_expr(indexprs, indrelid) AS expression
           FROM pg_index JOIN pg_class c ON (indexrelid = c.oid)
                JOIN pg_namespace ON (relnamespace = pg_namespace.oid)
                JOIN pg_roles ON (nspowner = pg_roles.oid)
                JOIN pg_am ON (relam = pg_am.oid)
           WHERE NOT indisprimary
                 AND (nspname = 'public' OR rolname <> 'postgres')
                 AND c.relname NOT IN (
                     SELECT conname FROM pg_constraint
                     WHERE contype = 'u')
           ORDER BY schema, 2, name"""

    def _from_catalog(self):
        """Initialize the dictionary of indexes by querying the catalogs"""
        for index in self.fetch():
            index.unqualify()
            sch, tbl, idx = index.key()
            if index.keycols == '0':
                del index.keycols
            self[(sch, tbl, idx)] = index

    def from_map(self, table, inindexes):
        """Initialize the dictionary of indexes by converting the input map

        :param table: table owning the indexes
        :param inindexes: YAML map defining the indexes
        """
        for i in inindexes.keys():
            idx = Index(schema=table.schema, table=table.name, name=i)
            val = inindexes[i]
            if 'columns' in val:
                idx.keycols = val['columns']
            elif 'expression' in val:
                idx.expression = val['expression']
            else:
                raise KeyError("Index '%s' is missing columns or expression"
                               % i)
            for attr in ['access_method', 'unique']:
                if attr in val:
                    setattr(idx, attr, val[attr])
            if not hasattr(idx, 'unique'):
                idx.unique = False
            self[(table.schema, table.name, i)] = idx

    def diff_map(self, inindexes):
        """Generate SQL to transform existing indexes

        :param inindexes: a YAML map defining the new indexes
        :return: list of SQL statements

        Compares the existing index definitions, as fetched from the
        catalogs, to the input map and generates SQL statements to
        transform the indexes accordingly.
        """
        stmts = []
        # check input indexes
        for (sch, tbl, idx) in inindexes.keys():
            inidx = inindexes[(sch, tbl, idx)]
            # does it exist in the database?
            if (sch, tbl, idx) not in self:
                # check for possible RENAME
                if hasattr(inidx, 'oldname'):
                    oldname = inidx.oldname
                    try:
                        stmts.append(self[(sch, tbl, oldname)].rename(
                                inidx.name))
                        del self[(sch, tbl, oldname)]
                    except KeyError, exc:
                        exc.args = ("Previous name '%s' for index '%s' "
                                  "not found" % (oldname, inidx.name), )
                        raise
                else:
                    # create new index
                    stmts.append(inidx.create())

        # check database indexes
        for (sch, tbl, idx) in self.keys():
            index = self[(sch, tbl, idx)]
            # if missing, drop it
            if (sch, tbl, idx) not in inindexes:
                stmts.append(index.drop())
            else:
                # compare index objects
                stmts.append(index.diff_map(inindexes[(sch, tbl, idx)]))

        return stmts



# -*- coding: utf-8 -*-
"""Test schemas"""

import unittest

from utils import PyrseasTestCase, fix_indent

CREATE_STMT = "CREATE SCHEMA s1"
DROP_STMT = "DROP SCHEMA IF EXISTS s1 CASCADE"
COMMENT_STMT = "COMMENT ON SCHEMA s1 IS 'Test schema s1'"


class SchemaToMapTestCase(PyrseasTestCase):
    """Test mapping of created schemas"""

    def test_map_schema(self):
        "Map a created schema"
        self.db.execute(DROP_STMT)
        dbmap = self.db.execute_and_map(CREATE_STMT)
        self.assertEqual(dbmap['schema s1'], {})

    def test_map_table_within_schema(self):
        "Map a schema and a table within it"
        self.db.execute(DROP_STMT)
        self.db.execute(CREATE_STMT)
        ddlstmt = "CREATE TABLE s1.t1 (c1 INTEGER, c2 TEXT)"
        expmap = {'table t1': {
                'columns': [{'c1': {'type': 'integer'}},
                            {'c2': {'type': 'text'}}]}}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema s1'], expmap)

    def test_map_schema_comment(self):
        "Map a schema comment"
        self.db.execute(DROP_STMT)
        self.db.execute(CREATE_STMT)
        dbmap = self.db.execute_and_map(COMMENT_STMT)
        self.assertEqual(dbmap['schema s1'], {'description': 'Test schema s1'})


class SchemaToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input schemas"""

    _schmap = {'schema s1': {'description': 'Test schema s1'}}

    def tearDown(self):
        self.db.execute_commit(DROP_STMT)
        self.db.close()

    def test_create_schema(self):
        "Create a schema that didn't exist"
        self.db.execute_commit(DROP_STMT)
        inmap = {'schema s1': {}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, [CREATE_STMT])

    def test_bad_schema_map(self):
        "Error creating a schema with a bad map"
        inmap = {'s1': {}}
        self.assertRaises(KeyError, self.db.process_map, inmap)

    def test_create_table_within_schema(self):
        "Create a new schema and a table within it"
        self.db.execute_commit(DROP_STMT)
        inmap = {'schema s1': {
                'table t1':
                    {'columns': [{'c1': {'type': 'integer'}},
                                 {'c2': {'type': 'text'}}]}}}
        dbsql = self.db.process_map(inmap)
        expsql = [CREATE_STMT,
                 "CREATE TABLE s1.t1 (c1 integer, c2 text)"]
        for i in range(len(expsql)):
            self.assertEqual(fix_indent(dbsql[i]), expsql[i])

    def test_drop_schema(self):
        "Drop an existing schema"
        self.db.execute(DROP_STMT)
        self.db.execute_commit(CREATE_STMT)
        dbsql = self.db.process_map({})
        self.assertEqual(dbsql, ["DROP SCHEMA s1"])

    def test_rename_schema(self):
        "Rename an existing schema"
        self.db.execute(DROP_STMT)
        self.db.execute_commit(CREATE_STMT)
        inmap = {'schema s2': {'oldname': 's1'}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["ALTER SCHEMA s1 RENAME TO s2"])

    def test_bad_rename_schema(self):
        "Error renaming a non-existing schema"
        self.db.execute(DROP_STMT)
        self.db.execute_commit(CREATE_STMT)
        inmap = {'schema s2': {'oldname': 's3'}}
        self.assertRaises(KeyError, self.db.process_map, inmap)

    def test_schema_with_comment(self):
        "Create a schema with a comment"
        self.db.execute_commit(DROP_STMT)
        dbsql = self.db.process_map(self._schmap)
        self.assertEqual(dbsql, [CREATE_STMT, COMMENT_STMT])

    def test_comment_on_schema(self):
        "Create a comment for an existing schema"
        self.db.execute(DROP_STMT)
        self.db.execute_commit(CREATE_STMT)
        dbsql = self.db.process_map(self._schmap)
        self.assertEqual(dbsql, [COMMENT_STMT])
        self.db.execute_commit("DROP SCHEMA s1")

    def test_drop_schema_comment(self):
        "Drop a comment on an existing schema"
        self.db.execute(DROP_STMT)
        self.db.execute(CREATE_STMT)
        self.db.execute_commit(COMMENT_STMT)
        dbsql = self.db.process_map({'schema s1': {}})
        self.assertEqual(dbsql, ["COMMENT ON SCHEMA s1 IS NULL"])
        self.db.execute_commit("DROP SCHEMA s1")

    def test_change_schema_comment(self):
        "Change existing comment on a schema"
        self.db.execute(DROP_STMT)
        self.db.execute(CREATE_STMT)
        self.db.execute_commit(COMMENT_STMT)
        inmap = {'schema s1': {'description': 'Changed schema s1'}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, [
                "COMMENT ON SCHEMA s1 IS 'Changed schema s1'"])
        self.db.execute_commit("DROP SCHEMA s1")


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(SchemaToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            SchemaToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""Utility functions and classes for testing Pyrseas"""

import os
from unittest import TestCase

from psycopg2 import connect
from psycopg2.extras import DictConnection
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

from pyrseas.dbconn import DbConnection
from pyrseas.database import Database


def fix_indent(stmt):
    "Fix specifications which are in a new line with indentation"
    return stmt.replace('   ', ' ').replace('  ', ' ').replace('\n ', ' '). \
        replace('( ', '(')


def new_std_map():
    "Return a standard public schema map with its description"
    return {'schema public': {'description': 'standard public schema'}}


def pgconnect(dbname, user, host, port):
    "Connect to a Postgres database using psycopg2"
    if host is None or host == '127.0.0.1' or host == 'localhost':
        host = ''
    else:
        host = 'host=%s ' % host
    if port is None or port == 5432:
        port = ''
    else:
        port = "port=%d " % port
    return connect("%s%sdbname=%s user=%s" % (
            host, port, dbname, user), connection_factory=DictConnection)


def pgexecute(dbconn, query):
    "Execute a query using a cursor"
    curs = dbconn.cursor()
    try:
        curs.execute(query)
    except:
        curs.close()
        dbconn.rollback()
        raise
    return curs


def pgexecute_auto(dbconn, query):
    "Execute a query using a cursor with autocommit enabled"
    isolation_level = dbconn.isolation_level
    dbconn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
    curs = pgexecute(dbconn, query)
    dbconn.set_isolation_level(isolation_level)
    return curs


TEST_DBNAME = os.environ.get("PYRSEAS_TEST_DB", 'pyrseas_testdb')
TEST_USER = os.environ.get("PYRSEAS_TEST_USER", os.getenv("USER"))
TEST_HOST = os.environ.get("PYRSEAS_TEST_HOST", None)
TEST_PORT = os.environ.get("PYRSEAS_TEST_PORT", None)
ADMIN_DB = os.environ.get("PYRSEAS_ADMIN_DB", 'postgres')
CREATE_DDL = "CREATE DATABASE %s TEMPLATE = template0"


class PostgresDb(object):
    """A PostgreSQL database connection

    This is separate from the one used by DbConnection, because the
    tests need to create and drop databases and other objects,
    independently.
    """
    def __init__(self, name, user, host, port):
        self.name = name
        self.conn = None
        self.user = user
        self.host = host
        self.port = port and int(port)
        self._version = 0

    def connect(self):
        """Connect to the database

        If we're not already connected we first connect to the admin
        database and see if the given database exists.  If it doesn't,
        we create and then connect to it.
        """
        if not self.conn:
            conn = pgconnect(ADMIN_DB, self.user, self.host, self.port)
            curs = pgexecute(conn,
                             "SELECT 1 FROM pg_database WHERE datname = '%s'" %
                             self.name)
            row = curs.fetchone()
            if not row:
                curs.close()
                curs = pgexecute_auto(conn, CREATE_DDL % self.name)
                curs.close()
            conn.close()
            self.conn = pgconnect(self.name, self.user, self.host, self.port)
            curs = pgexecute(self.conn, "SHOW server_version_num")
            self._version = int(curs.fetchone()[0])

    def close(self):
        "Close the connection if still open"
        if not self.conn:
            return ValueError
        self.conn.close()

    @property
    def version(self):
        return self._version

    def create(self):
        "Drop the database if it exists and re-create it"
        conn = pgconnect(ADMIN_DB, self.user, self.host, self.port)
        curs = pgexecute_auto(conn, "DROP DATABASE IF EXISTS %s" % self.name)
        curs = pgexecute_auto(conn, CREATE_DDL % self.name)
        curs.close()
        conn.close()

    def clear(self):
        "Drop tables and other objects"
        curs = pgexecute(
            self.conn,
            """SELECT nspname, relname, relkind FROM pg_class
                      JOIN pg_namespace ON (relnamespace = pg_namespace.oid)
                      JOIN pg_roles ON (nspowner = pg_roles.oid)
               WHERE relkind in ('r', 'S', 'v')
                     AND (nspname = 'public' OR rolname <> 'postgres')
               ORDER BY relkind DESC""")
        objs = curs.fetchall()
        curs.close()
        self.conn.rollback()
        for obj in objs:
            if obj['relkind'] == 'r':
                self.execute("DROP TABLE IF EXISTS %s.%s CASCADE" % (
                        obj[0], obj[1]))
            elif obj['relkind'] == 'S':
                self.execute("DROP SEQUENCE %s.%s CASCADE" % (obj[0], obj[1]))
            elif obj['relkind'] == 'v':
                self.execute("DROP VIEW %s.%s CASCADE" % (obj[0], obj[1]))
        self.conn.commit()
        curs = pgexecute(
            self.conn,
            """SELECT nspname, proname, pg_get_function_arguments(p.oid) AS
                      args
               FROM pg_proc p JOIN pg_namespace n ON (pronamespace = n.oid)
               WHERE (nspname != 'pg_catalog'
                     AND nspname != 'information_schema')""")
        funcs = curs.fetchall()
        curs.close()
        self.conn.rollback()
        for func in funcs:
            self.execute("DROP FUNCTION IF EXISTS %s.%s (%s) CASCADE" % (
                    func[0], func[1], func[2]))
        self.conn.commit()
        if self.version < 90000:
            if self.is_plpgsql_installed():
                self.execute_commit("DROP LANGUAGE plpgsql")
        curs = pgexecute(
            self.conn,
            """SELECT nspname, typname, typtype FROM pg_type t
                      JOIN pg_namespace n ON (typnamespace = n.oid)
               WHERE (nspname != 'pg_catalog'
                     AND nspname != 'information_schema')""")
        types = curs.fetchall()
        curs.close()
        self.conn.rollback()
        for typ in types:
            if typ['typtype'] == 'd':
                self.execute("DROP DOMAIN IF EXISTS %s.%s CASCADE" % (
                        typ[0], typ[1]))
            elif typ['typtype'] == 'e':
                self.execute("DROP TYPE IF EXISTS %s.%s CASCADE" % (
                        typ[0], typ[1]))
        self.conn.commit()

    def drop(self):
        "Drop the database"
        conn = pgconnect(ADMIN_DB, self.user, self.host, self.port)
        curs = pgexecute_auto(conn, "DROP DATABASE %s" % self.name)
        curs.close()
        conn.close()

    def execute(self, stmt):
        "Execute a DDL statement"
        curs = pgexecute(self.conn, stmt)
        curs.close()

    def execute_commit(self, stmt):
        "Execute a DDL statement and commit"
        self.execute(stmt)
        self.conn.commit()

    def is_plpgsql_installed(self):
        "Is PL/pgSQL installed?"
        curs = pgexecute(self.conn,
                         "SELECT 1 FROM pg_language WHERE lanname = 'plpgsql'")
        row = curs.fetchone()
        curs.close()
        return row and True

    def execute_and_map(self, ddlstmt):
        "Execute a DDL statement, commit, and return a map of the database"
        self.execute(ddlstmt)
        self.conn.commit()
        db = Database(DbConnection(self.name, self.user, self.host, self.port))
        return db.to_map()

    def process_map(self, input_map):
        """Process an input map and return the SQL statements necessary to
        convert the database to match the map."""
        db = Database(DbConnection(self.name, self.user, self.host, self.port))
        stmts = db.diff_map(input_map)
        return stmts


class PyrseasTestCase(TestCase):
    """Base class for most test cases"""

    def setUp(self):
        self.db = PostgresDb(TEST_DBNAME, TEST_USER, TEST_HOST, TEST_PORT)
        self.db.connect()
        self.db.clear()

    def tearDown(self):
        self.db.close()



# -*- coding: utf-8 -*-
"""
    pyrseas.database
    ~~~~~~~~~~~~~~~~

    A `Database` is initialized with a DbConnection object.  It
    consists of one or two `Dicts` objects, each holding various
    dictionary objects.  The `db` Dicts object defines the database
    schemas, including their tables and other objects, by querying the
    system catalogs.  The `ndb` Dicts object defines the schemas based
    on the `input_map` supplied to the `from_map` method.
"""
from pyrseas.dbobject.language import LanguageDict
from pyrseas.dbobject.schema import SchemaDict
from pyrseas.dbobject.dbtype import TypeDict
from pyrseas.dbobject.table import ClassDict
from pyrseas.dbobject.column import ColumnDict
from pyrseas.dbobject.constraint import ConstraintDict
from pyrseas.dbobject.index import IndexDict
from pyrseas.dbobject.function import ProcDict
from pyrseas.dbobject.rule import RuleDict
from pyrseas.dbobject.trigger import TriggerDict


def flatten(lst):
    "Flatten a list possibly containing lists to a single list"
    for elem in lst:
        if isinstance(elem, list) and not isinstance(elem, basestring):
            for subelem in flatten(elem):
                yield subelem
        else:
            yield elem


class Database(object):
    """A database definition, from its catalogs and/or a YAML spec."""

    class Dicts(object):
        """A holder for dictionaries (maps) describing a database"""

        def __init__(self, dbconn=None):
            """Initialize the various DbObjectDict-derived dictionaries

            :param dbconn: a DbConnection object
            """
            self.languages = LanguageDict(dbconn)
            self.schemas = SchemaDict(dbconn)
            self.types = TypeDict(dbconn)
            self.tables = ClassDict(dbconn)
            self.columns = ColumnDict(dbconn)
            self.constraints = ConstraintDict(dbconn)
            self.indexes = IndexDict(dbconn)
            self.functions = ProcDict(dbconn)
            self.rules = RuleDict(dbconn)
            self.triggers = TriggerDict(dbconn)

    def __init__(self, dbconn):
        """Initialize the database

        :param dbconn: a DbConnection object
        """
        self.dbconn = dbconn
        self.db = None

    def _link_refs(self, db):
        """Link related objects"""
        db.languages.link_refs(db.functions)
        db.schemas.link_refs(db.types, db.tables, db.functions)
        db.tables.link_refs(db.columns, db.constraints, db.indexes,
                            db.rules, db.triggers)
        db.types.link_refs(db.constraints)

    def from_catalog(self):
        """Populate the database objects by querying the catalogs

        The `db` holder is populated by various DbObjectDict-derived
        classes by querying the catalogs. The objects in the
        dictionary are then linked to related objects, e.g., columns
        are linked to the tables they belong.
        """
        self.db = self.Dicts(self.dbconn)
        if self.dbconn.conn:
            self.dbconn.conn.close()
        self._link_refs(self.db)

    def from_map(self, input_map):
        """Populate the new database objects from the input map

        :param input_map: a YAML map defining the new database

        The `ndb` holder is populated by various DbObjectDict-derived
        classes by traversing the YAML input map. The objects in the
        dictionary are then linked to related objects, e.g., columns
        are linked to the tables they belong.
        """
        self.ndb = self.Dicts()
        input_schemas = {}
        input_langs = {}
        for key in input_map.keys():
            if key.startswith('schema '):
                input_schemas.update({key: input_map[key]})
            elif key.startswith('language '):
                input_langs.update({key: input_map[key]})
            else:
                raise KeyError("Expected typed object, found '%s'" % key)
        self.ndb.languages.from_map(input_langs)
        self.ndb.schemas.from_map(input_schemas, self.ndb)
        self._link_refs(self.ndb)

    def to_map(self):
        """Convert the db maps to a single hierarchy suitable for YAML

        :return: a YAML-suitable dictionary (without Python objects)
        """
        if not self.db:
            self.from_catalog()
        dbmap = self.db.languages.to_map()
        dbmap.update(self.db.schemas.to_map())
        return dbmap

    def diff_map(self, input_map):
        """Generate SQL to transform an existing database

        :param input_map: a YAML map defining the new database
        :return: list of SQL statements

        Compares the existing database definition, as fetched from the
        catalogs, to the input YAML map and generates SQL statements
        to transform the database into the one represented by the
        input.
        """
        if not self.db:
            self.from_catalog()
        self.from_map(input_map)
        stmts = self.db.languages.diff_map(self.ndb.languages,
                                           self.dbconn.version)
        stmts.append(self.db.schemas.diff_map(self.ndb.schemas))
        stmts.append(self.db.types.diff_map(self.ndb.types))
        stmts.append(self.db.functions.diff_map(self.ndb.functions))
        stmts.append(self.db.tables.diff_map(self.ndb.tables))
        stmts.append(self.db.constraints.diff_map(self.ndb.constraints))
        stmts.append(self.db.indexes.diff_map(self.ndb.indexes))
        stmts.append(self.db.columns.diff_map(self.ndb.columns))
        stmts.append(self.db.triggers.diff_map(self.ndb.triggers))
        stmts.append(self.db.rules.diff_map(self.ndb.rules))
        stmts.append(self.db.functions._drop())
        stmts.append(self.db.types._drop())
        stmts.append(self.db.schemas._drop())
        stmts.append(self.db.languages._drop())
        return [s for s in flatten(stmts)]



# -*- coding: utf-8 -*-
"""Test constraints"""

import unittest

from utils import PyrseasTestCase, fix_indent, new_std_map

DROP_STMT = "DROP TABLE IF EXISTS t1"


class CheckConstraintToMapTestCase(PyrseasTestCase):
    """Test mapping of created CHECK constraints"""

    def test_check_constraint_1(self):
        "Map a table with a CHECK constraint"
        ddlstmt = "CREATE TABLE t1 (c1 INTEGER, c2 SMALLINT CHECK (c2 < 1000))"
        expmap = {'columns': [{'c1': {'type': 'integer'}},
                              {'c2': {'type': 'smallint'}}],
                  'check_constraints': {'t1_c2_check': {
                    'columns': ['c2'],
                    'expression': '(c2 < 1000)'}}}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)

    def test_check_constraint_2(self):
        "Map a table with a two-column, named CHECK constraint"
        ddlstmt = "CREATE TABLE t1 (c1 INTEGER, c2 INTEGER, " \
            "CONSTRAINT t1_check_ratio CHECK (c2 * 100 / c1 <= 50))"
        expmap = {'columns': [{'c1': {'type': 'integer'}},
                              {'c2': {'type': 'integer'}}],
                  'check_constraints': {'t1_check_ratio': {
                    'columns': ['c2', 'c1'],
                    'expression': '(((c2 * 100) / c1) <= 50)'}}}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)


class CheckConstraintToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input CHECK constraints"""

    def test_create_w_check_constraint(self):
        "Create new table with a single column CHECK constraint"
        self.db.execute_commit(DROP_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'check_constraints': {'t1_c1_check': {
                            'columns': ['c1'],
                            'expression': 'c1 > 0 and c1 < 1000000'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "CREATE TABLE t1 (c1 integer, c2 text)")
        self.assertEqual(fix_indent(dbsql[1]),
                         "ALTER TABLE t1 ADD CONSTRAINT t1_c1_check "
                         "CHECK (c1 > 0 and c1 < 1000000)")

    def test_add_check_constraint(self):
        "Add a two-column CHECK constraint to an existing table"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER NOT NULL, "
                        "c2 INTEGER NOT NULL, c3 TEXT)")
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                'columns': [
                    {'c1': {'type': 'integer', 'not_null': True}},
                    {'c2': {'type': 'integer', 'not_null': True}},
                    {'c3': {'type': 'text'}}],
                'check_constraints': {'t1_check_2_1': {
                        'columns': ['c2', 'c1'],
                        'expression': 'c2 != c1'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "ALTER TABLE t1 ADD CONSTRAINT t1_check_2_1 "
                         "CHECK (c2 != c1)")


class PrimaryKeyToMapTestCase(PyrseasTestCase):
    """Test mapping of created PRIMARY KEYs"""

    map_pkey1 = {'columns': [{'c1': {'type': 'integer', 'not_null': True}},
                             {'c2': {'type': 'text'}}],
                 'primary_key': {'t1_pkey': {'columns': ['c1'],
                                             'access_method': 'btree'}}}
    map_pkey2 = {'columns': [
            {'c1': {'type': 'integer', 'not_null': True}},
            {'c2': {'type': 'character(5)', 'not_null': True}},
            {'c3': {'type': 'text'}}],
                 'primary_key': {'t1_pkey': {'columns': ['c2', 'c1'],
                                             'access_method': 'btree'}}}

    map_pkey3 = {'columns': [{'c1': {'type': 'integer', 'not_null': True}},
                             {'c2': {'type': 'text'}}],
                 'primary_key': {'t1_prim_key': {'columns': ['c1'],
                                                 'access_method': 'btree'}}}

    def test_primary_key_1(self):
        "Map a table with a single-column primary key"
        ddlstmt = "CREATE TABLE t1 (c1 INTEGER PRIMARY KEY, c2 TEXT)"
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_pkey1)

    def test_primary_key_2(self):
        "Map a table with a single-column primary key, table-level constraint"
        ddlstmt = "CREATE TABLE t1 (c1 INTEGER, c2 TEXT, PRIMARY KEY (c1))"
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_pkey1)

    def test_primary_key_3(self):
        "Map a table with two-column primary key, atypical order"
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 CHAR(5), c3 TEXT,
                                      PRIMARY KEY (c2, c1))"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_pkey2)

    def test_primary_key_4(self):
        "Map a table with a named primary key constraint"
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 TEXT,
                                  CONSTRAINT t1_prim_key PRIMARY KEY (c1))"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_pkey3)

    def test_primary_key_5(self):
        "Map a table with a named primary key, column level constraint"
        ddlstmt = """CREATE TABLE t1 (
                            c1 INTEGER CONSTRAINT t1_prim_key PRIMARY KEY,
                            c2 TEXT)"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_pkey3)


class PrimaryKeyToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input PRIMARY KEYs"""

    def test_create_with_primary_key(self):
        "Create new table with single column primary key"
        self.db.execute_commit(DROP_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'text'}},
                                {'c2': {'type': 'integer'}}],
                    'primary_key': {'t1_pkey': {
                            'columns': ['c2'],
                            'access_method': 'btree'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "CREATE TABLE t1 (c1 text, c2 integer)")
        self.assertEqual(fix_indent(dbsql[1]),
                         "ALTER TABLE t1 ADD CONSTRAINT t1_pkey "
                         "PRIMARY KEY (c2)")

    def test_add_primary_key(self):
        "Add a two-column primary key to an existing table"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER NOT NULL, "
                        "c2 INTEGER NOT NULL, c3 TEXT)")
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [
                        {'c1': {'type': 'integer', 'not_null': True}},
                        {'c2': {'type': 'integer', 'not_null': True}},
                        {'c3': {'type': 'text'}}],
                    'primary_key': {'t1_pkey': {
                            'columns': ['c1', 'c2'],
                            'access_method': 'btree'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "ALTER TABLE t1 ADD CONSTRAINT t1_pkey "
                         "PRIMARY KEY (c1, c2)")

    def test_drop_primary_key(self):
        "Drop a primary key on an existing table"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER NOT NULL "
                               "PRIMARY KEY, c2 TEXT)")
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer', 'not_null': True},
                                 'c2': {'type': 'text'}}]}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["ALTER TABLE t1 DROP CONSTRAINT t1_pkey"])


class ForeignKeyToMapTestCase(PyrseasTestCase):
    """Test mapping of created FOREIGN KEYs"""

    map_fkey1 = {'columns': [{'c1': {'type': 'integer'}},
                             {'c2': {'type': 'integer'}},
                             {'c3': {'type': 'text'}}],
                 'foreign_keys': {'t1_c2_fkey': {
                    'columns': ['c2'],
                    'references': {'schema': 'public', 'table': 't2',
                                   'columns': ['pc1']}}}}

    map_fkey2 = {'columns': [{'c1': {'type': 'integer'}},
                             {'c2': {'type': 'character(5)'}},
                             {'c3': {'type': 'integer'}},
                             {'c4': {'type': 'date'}},
                             {'c5': {'type': 'text'}}],
                 'foreign_keys': {'t1_c2_fkey': {
                    'columns': ['c2', 'c3', 'c4'],
                    'references': {'schema': 'public', 'table': 't2',
                                   'columns': ['pc2', 'pc1', 'pc3']}}}}

    map_fkey3 = {'columns': [{'c1': {'type': 'integer'}},
                             {'c2': {'type': 'character(5)'}},
                             {'c3': {'type': 'integer'}},
                             {'c4': {'type': 'date'}},
                             {'c5': {'type': 'text'}}],
                 'foreign_keys': {'t1_fgn_key': {
                    'columns': ['c2', 'c3', 'c4'],
                    'references': {'schema': 'public', 'table': 't2',
                                   'columns': ['pc2', 'pc1', 'pc3']}}}}

    map_fkey4 = {'columns': [{'c1': {'type': 'integer', 'not_null': True}},
                             {'c2': {'type': 'character(5)',
                                     'not_null': True}},
                             {'c3': {'type': 'integer'}},
                             {'c4': {'type': 'date'}},
                             {'c5': {'type': 'text'}}],
                 'primary_key': {'t1_prim_key': {'columns': ['c1', 'c2'],
                                                 'access_method': 'btree'}},
                 'foreign_keys': {'t1_fgn_key1': {
                    'columns': ['c2', 'c3', 'c4'],
                    'references': {'schema': 'public', 'table': 't2',
                                   'columns': ['pc2', 'pc1', 'pc3']}},
                                  't1_fgn_key2': {
                    'columns': ['c2'],
                    'references': {'schema': 'public', 'table': 't3',
                                   'columns': ['qc1']}}}}

    def test_foreign_key_1(self):
        "Map a table with a single-column foreign key on another table"
        self.db.execute("CREATE TABLE t2 (pc1 INTEGER PRIMARY KEY, pc2 TEXT)")
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER,
                          c2 INTEGER REFERENCES t2 (pc1), c3 TEXT)"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_fkey1)

    def test_foreign_key_2(self):
        "Map a table with a single-column foreign key, table level constraint"
        self.db.execute("CREATE TABLE t2 (pc1 INTEGER PRIMARY KEY, pc2 TEXT)")
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 INTEGER, c3 TEXT,
                            FOREIGN KEY (c2) REFERENCES t2 (pc1))"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_fkey1)

    def test_foreign_key_3(self):
        "Map a table with a three-column foreign key"
        self.db.execute("""CREATE TABLE t2 (pc1 INTEGER, pc2 CHAR(5), pc3 DATE,
                                  pc4 TEXT, PRIMARY KEY (pc2, pc1, pc3))""")
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 CHAR(5), c3 INTEGER,
                            c4 DATE, c5 TEXT,
                            FOREIGN KEY (c2, c3, c4)
                                REFERENCES t2 (pc2, pc1, pc3))"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_fkey2)

    def test_foreign_key_4(self):
        "Map a table with a named, three-column foreign key"
        self.db.execute("""CREATE TABLE t2 (pc1 INTEGER, pc2 CHAR(5), pc3 DATE,
                                  pc4 TEXT, PRIMARY KEY (pc2, pc1, pc3))""")
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 CHAR(5), c3 INTEGER,
                            c4 DATE, c5 TEXT,
                            CONSTRAINT t1_fgn_key FOREIGN KEY (c2, c3, c4)
                                REFERENCES t2 (pc2, pc1, pc3))"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_fkey3)

    def test_foreign_key_5(self):
        "Map a table with a primary key and two foreign keys"
        self.db.execute("""CREATE TABLE t2 (pc1 INTEGER, pc2 CHAR(5), pc3 DATE,
                                  pc4 TEXT, PRIMARY KEY (pc2, pc1, pc3))""")
        self.db.execute("CREATE TABLE t3 (qc1 CHAR(5) PRIMARY KEY, qc2 text)")
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 CHAR(5), c3 INTEGER,
                            c4 DATE, c5 TEXT,
                            CONSTRAINT t1_prim_key PRIMARY KEY (c1, c2),
                            CONSTRAINT t1_fgn_key1 FOREIGN KEY (c2, c3, c4)
                                REFERENCES t2 (pc2, pc1, pc3),
                            CONSTRAINT t1_fgn_key2 FOREIGN KEY (c2)
                                REFERENCES t3 (qc1))"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_fkey4)

    def test_foreign_key_actions(self):
        "Map a table with foreign key ON UPDATE/ON DELETE actions"
        self.db.execute("CREATE TABLE t2 (pc1 INTEGER PRIMARY KEY, pc2 TEXT)")
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 INTEGER, c3 TEXT,
                            FOREIGN KEY (c2) REFERENCES t2 (pc1)
                                ON UPDATE RESTRICT ON DELETE SET NULL)"""
        expmap = {'columns': [{'c1': {'type': 'integer'}},
                              {'c2': {'type': 'integer'}},
                              {'c3': {'type': 'text'}}],
                  'foreign_keys': {'t1_c2_fkey': {
                    'columns': ['c2'],
                    'on_update': 'restrict',
                    'on_delete': 'set null',
                    'references': {'schema': 'public', 'table': 't2',
                                   'columns': ['pc1']}}}}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)

    def test_cross_schema_foreign_key(self):
        "Map a table with a foreign key on a table in another schema"
        self.db.execute("DROP SCHEMA IF EXISTS s1 CASCADE")
        self.db.execute("CREATE SCHEMA s1")
        self.db.execute("CREATE TABLE t2 (pc1 INTEGER PRIMARY KEY, pc2 TEXT)")
        ddlstmt = """CREATE TABLE s1.t1 (c1 INTEGER PRIMARY KEY,
                          c2 INTEGER REFERENCES t2 (pc1), c3 TEXT)"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.db.execute_commit("DROP SCHEMA s1 CASCADE")
        t2map = {'columns': [{'pc1': {'type': 'integer', 'not_null': True}},
                             {'pc2': {'type': 'text'}}],
                 'primary_key': {'t2_pkey': {
                    'columns': ['pc1'], 'access_method': 'btree'}}}
        t1map = {'table t1': {
                'columns': [{'c1': {'type': 'integer', 'not_null': True}},
                            {'c2': {'type': 'integer'}},
                            {'c3': {'type': 'text'}}],
                'primary_key': {'t1_pkey': {
                        'columns': ['c1'],
                        'access_method': 'btree'}},
                'foreign_keys': {'t1_c2_fkey': {
                        'columns': ['c2'],
                        'references': {'schema': 'public', 'table': 't2',
                                       'columns': ['pc1']}}}}}
        self.assertEqual(dbmap['schema public']['table t2'], t2map)
        self.assertEqual(dbmap['schema s1'], t1map)

    def test_multiple_foreign_key(self):
        "Map a table with its primary key referenced by two others"
        self.db.execute("CREATE TABLE t1 (pc1 integer PRIMARY KEY, pc2 text)")
        ddlstmt = """CREATE TABLE t2 (c1 integer,
                          c2 integer REFERENCES t1 (pc1), c3 text,
                          c4 integer REFERENCES t1 (pc1))"""
        dbmap = self.db.execute_and_map(ddlstmt)
        t1map = {'columns': [{'pc1': {'type': 'integer', 'not_null': True}},
                             {'pc2': {'type': 'text'}}],
                 'primary_key': {'t1_pkey': {
                    'columns': ['pc1'], 'access_method': 'btree'}}}
        t2map = {'columns': [{'c1': {'type': 'integer'}},
                             {'c2': {'type': 'integer'}},
                             {'c3': {'type': 'text'}},
                             {'c4': {'type': 'integer'}}],
                 'foreign_keys': {'t2_c2_fkey': {
                    'columns': ['c2'],
                    'references': {'schema': 'public', 'table': 't1',
                                   'columns': ['pc1']}},
                                  't2_c4_fkey': {
                    'columns': ['c4'],
                    'references': {'schema': 'public', 'table': 't1',
                                   'columns': ['pc1']}}}}
        self.assertEqual(dbmap['schema public']['table t1'], t1map)
        self.assertEqual(dbmap['schema public']['table t2'], t2map)

    def test_foreign_key_dropped_column(self):
        "Map a table with a foreign key after a column has been dropped"
        self.db.execute("CREATE TABLE t1 (pc1 integer PRIMARY KEY, pc2 text)")
        self.db.execute("CREATE TABLE t2 (c1 integer, c2 text, c3 smallint, "
                       "c4 integer REFERENCES t1 (pc1))")
        ddlstmt = "ALTER TABLE t2 DROP COLUMN c3"
        dbmap = self.db.execute_and_map(ddlstmt)
        t2map = {'columns': [{'c1': {'type': 'integer'}},
                             {'c2': {'type': 'text'}},
                             {'c4': {'type': 'integer'}}],
                 'foreign_keys': {'t2_c4_fkey': {
                    'columns': ['c4'],
                    'references': {'schema': 'public', 'table': 't1',
                                   'columns': ['pc1']}}}}
        self.assertEqual(dbmap['schema public']['table t2'], t2map)


class ForeignKeyToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input FOREIGN KEYs"""

    def test_create_with_foreign_key(self):
        "Create a table with a foreign key constraint"
        self.db.execute_commit(DROP_STMT + ", t2")
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c11': {'type': 'integer'}},
                                {'c12': {'type': 'text'}}]},
                                   'table t2': {
                    'columns': [{'c21': {'type': 'integer'}},
                                {'c22': {'type': 'text'}},
                                {'c23': {'type': 'integer'}}],
                    'foreign_keys': {'t2_c23_fkey': {
                            'columns': ['c23'],
                            'references': {'columns': ['c11'],
                                           'table': 't1'}}}}})
        dbsql = self.db.process_map(inmap)
        # can't control which table will be created first
        crt1 = 0
        crt2 = 1
        if 't1' in dbsql[1]:
            crt1 = 1
            crt2 = 0
        self.assertEqual(fix_indent(dbsql[crt1]),
                             "CREATE TABLE t1 (c11 integer, c12 text)")
        self.assertEqual(fix_indent(dbsql[crt2]),
                         "CREATE TABLE t2 (c21 integer, c22 text, "
                         "c23 integer)")
        self.assertEqual(fix_indent(dbsql[2]),
                         "ALTER TABLE t2 ADD CONSTRAINT t2_c23_fkey "
                         "FOREIGN KEY (c23) REFERENCES t1 (c11)")

    def test_add_foreign_key(self):
        "Add a two-column foreign key to an existing table"
        self.db.execute(DROP_STMT + ", t2")
        self.db.execute("CREATE TABLE t1 (c11 INTEGER NOT NULL, "
                        "c12 INTEGER NOT NULL, c13 TEXT, "
                        "PRIMARY KEY (c11, c12))")
        self.db.execute_commit("CREATE TABLE t2 (c21 INTEGER NOT NULL, "
                               "c22 TEXT, c23 INTEGER, c24 INTEGER, "
                               "PRIMARY KEY (c21))")
        inmap = new_std_map()
        inmap['schema public'].update({
                'table t1': {'columns': [
                        {'c11': {'type': 'integer', 'not_null': True}},
                        {'c12': {'type': 'integer', 'not_null': True}},
                        {'c13': {'type': 'text'}}],
                             'primary_key': {'t1_pkey': {
                            'columns': ['c11', 'c12'],
                            'access_method': 'btree'}}},
                'table t2': {'columns': [
                        {'c21': {'type': 'integer', 'not_null': True}},
                        {'c22': {'type': 'text'}},
                        {'c23': {'type': 'integer'}},
                        {'c24': {'type': 'integer'}}],
                             'primary_key': {'t2_pkey': {
                            'columns': ['c21'],
                            'access_method': 'btree'}},
                             'foreign_keys': {'t2_c23_fkey': {
                            'columns': ['c23', 'c24'],
                            'references': {'columns': ['c11', 'c12'],
                                           'table': 't1'}}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "ALTER TABLE t2 ADD CONSTRAINT t2_c23_fkey "
                         "FOREIGN KEY (c23, c24) REFERENCES t1 (c11, c12)")

    def test_drop_foreign_key(self):
        "Drop a foreign key on an existing table"
        self.db.execute(DROP_STMT + ", t2")
        self.db.execute("CREATE TABLE t1 (c11 INTEGER NOT NULL, c12 TEXT, "
                        "PRIMARY KEY (c11))")
        self.db.execute_commit("CREATE TABLE t2 (c21 INTEGER NOT NULL "
                               "PRIMARY KEY, c22 INTEGER NOT NULL "
                               "REFERENCES t1 (c11), c23 TEXT)")
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c11': {'type': 'integer', 'not_null': True}},
                                {'c12': {'type': 'text'}}],
                    'primary_key': {'t1_pkey': {'columns': ['c11']}}},
                                       'table t2': {
                    'columns': [{'c21': {'type': 'integer', 'not_null': True}},
                                {'c22': {'type': 'integer', 'not_null': True}},
                                {'c23': {'type': 'text'}}],
                    'primary_key': {'t2_pkey': {'columns': ['c21']}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["ALTER TABLE t2 DROP CONSTRAINT t2_c22_fkey"])

    def test_create_foreign_key_actions(self):
        "Create a table with foreign key ON UPDATE/ON DELETE actions"
        self.db.execute_commit(DROP_STMT + ", t2")
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c11': {'type': 'integer'}},
                                {'c12': {'type': 'text'}}]},
                                   'table t2': {
                    'columns': [{'c21': {'type': 'integer'}},
                                {'c22': {'type': 'text'}},
                                {'c23': {'type': 'integer'}}],
                    'foreign_keys': {'t2_c23_fkey': {
                            'columns': ['c23'],
                            'on_update': 'cascade',
                            'on_delete': 'set default',
                            'references': {'columns': ['c11'],
                                           'table': 't1'}}}}})
        dbsql = self.db.process_map(inmap)
        # won't check CREATE TABLEs explicitly here (see first test instead)
        self.assertEqual(fix_indent(dbsql[2]),
                         "ALTER TABLE t2 ADD CONSTRAINT t2_c23_fkey "
                         "FOREIGN KEY (c23) REFERENCES t1 (c11) "
                         "ON UPDATE CASCADE ON DELETE SET DEFAULT")


class UniqueConstraintToMapTestCase(PyrseasTestCase):
    """Test mapping of created UNIQUE constraints"""

    map_unique1 = {'columns': [{'c1': {'type': 'integer'}},
                               {'c2': {'type': 'text'}}],
                   'unique_constraints': {'t1_c1_key': {
                'columns': ['c1'], 'access_method': 'btree'}}}

    map_unique2 = {'columns': [{'c1': {'type': 'integer'}},
                               {'c2': {'type': 'character(5)'}},
                               {'c3': {'type': 'text'}}],
                   'unique_constraints': {'t1_c1_c2_key': {
                'columns': ['c1', 'c2'], 'access_method': 'btree'}}}

    map_unique3 = {'columns': [{'c1': {'type': 'integer'}},
                               {'c2': {'type': 'text'}}],
                   'unique_constraints': {'t1_unique_key': {
                'columns': ['c1'], 'access_method': 'btree'}}}

    def test_unique_1(self):
        "Map a table with a single-column unique constraint"
        ddlstmt = "CREATE TABLE t1 (c1 INTEGER UNIQUE, c2 TEXT)"
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_unique1)

    def test_unique_2(self):
        "Map a table with a single-column unique constraint, table level"
        ddlstmt = "CREATE TABLE t1 (c1 INTEGER, c2 TEXT, UNIQUE (c1))"
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_unique1)

    def test_unique_3(self):
        "Map a table with a two-column unique constraint"
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 CHAR(5), c3 TEXT,
                                    UNIQUE (c1, c2))"""
        dbmap = self.db.execute_and_map(ddlstmt)
        if self.db.version < 90000:
            self.map_unique2.update({'unique_constraints': {'t1_c1_key': {
                'columns': ['c1', 'c2'], 'access_method': 'btree'}}})
        self.assertEqual(dbmap['schema public']['table t1'], self.map_unique2)

    def test_unique_4(self):
        "Map a table with a named unique constraint"
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 TEXT,
                                  CONSTRAINT t1_unique_key UNIQUE (c1))"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_unique3)

    def test_unique_5(self):
        "Map a table with a named unique constraint, column level"
        ddlstmt = """CREATE TABLE t1 (
                            c1 INTEGER CONSTRAINT t1_unique_key UNIQUE,
                            c2 TEXT)"""
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], self.map_unique3)


class UniqueConstraintToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input UNIQUE constraints"""

    def test_create_w_unique_constraint(self):
        "Create new table with a single column unique constraint"
        self.db.execute_commit(DROP_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'unique_constraints': {'t1_c1_key': {
                            'columns': ['c1'],
                            'access_method': 'btree'}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "CREATE TABLE t1 (c1 integer, c2 text)")
        self.assertEqual(fix_indent(dbsql[1]),
                         "ALTER TABLE t1 ADD CONSTRAINT t1_c1_key "
                         "UNIQUE (c1)")

    def test_add_unique_constraint(self):
        "Add a two-column unique constraint to an existing table"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER NOT NULL, "
                        "c2 INTEGER NOT NULL, c3 TEXT)")
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                'columns': [
                    {'c1': {'type': 'integer', 'not_null': True}},
                    {'c2': {'type': 'integer', 'not_null': True}},
                    {'c3': {'type': 'text'}}],
                'unique_constraints': {'t1_c2_key': {
                        'columns': ['c2', 'c1'],
                        'unique': True}}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "ALTER TABLE t1 ADD CONSTRAINT t1_c2_key "
                         "UNIQUE (c2, c1)")

    def test_drop_unique_constraint(self):
        "Drop a unique constraint on an existing table"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER NOT NULL UNIQUE, "
                               "c2 TEXT)")
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer', 'not_null': True},
                                 'c2': {'type': 'text'}}]}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["ALTER TABLE t1 DROP CONSTRAINT t1_c1_key"])


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(
        CheckConstraintToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            CheckConstraintToSqlTestCase))
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            PrimaryKeyToMapTestCase))
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            PrimaryKeyToSqlTestCase))
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            ForeignKeyToMapTestCase))
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            ForeignKeyToSqlTestCase))
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            UniqueConstraintToMapTestCase))
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            UniqueConstraintToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
    pyrseas.function
    ~~~~~~~~~~~~~~~~

    This module defines four classes: Proc derived from
    DbSchemaObject, Function and Aggregate derived from Proc, and
    FunctionDict derived from DbObjectDict.
"""
from pyrseas.dbobject import DbObjectDict, DbSchemaObject


VOLATILITY_TYPES = {'i': 'immutable', 's': 'stable', 'v': 'volatile'}


class Proc(DbSchemaObject):
    """A procedure such as a FUNCTION or an AGGREGATE"""

    keylist = ['schema', 'name', 'arguments']

    def extern_key(self):
        """Return the key to be used in external maps for this function

        :return: string
        """
        return '%s %s(%s)' % (self.objtype.lower(), self.name, self.arguments)

    def identifier(self):
        """Return a full identifier for a function object

        :return: string
        """
        return "%s(%s)" % (self.qualname(), self.arguments)


class Function(Proc):
    """A procedural language function"""

    objtype = "FUNCTION"

    def to_map(self):
        """Convert a function to a YAML-suitable format

        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        if self.volatility == 'v':
            del dct['volatility']
        else:
            dct['volatility'] = VOLATILITY_TYPES[self.volatility]
        if hasattr(self, 'dependent_table'):
            del dct['dependent_table']
        return {self.extern_key(): dct}

    def create(self, newsrc=None):
        """Return SQL statements to CREATE or REPLACE the function

        :param newsrc: new source for a changed function
        :return: SQL statements
        """
        stmts = []
        if hasattr(self, 'dependent_table'):
            stmts.append(self.dependent_table.create())
        src = newsrc or self.source
        volat = strict = secdef = ''
        if hasattr(self, 'volatility'):
            volat = ' ' + VOLATILITY_TYPES[self.volatility].upper()
        if hasattr(self, 'strict') and self.strict:
            strict = ' STRICT'
        if hasattr(self, 'security_definer') and self.security_definer:
            secdef = ' SECURITY DEFINER'
        stmts.append("CREATE%s FUNCTION %s(%s) RETURNS %s\n    LANGUAGE %s"
                     "\n    AS $_$%s$_$%s%s%s" % (
                newsrc and " OR REPLACE" or '', self.qualname(),
                self.arguments, self.returns, self.language, src,
                volat, strict, secdef))
        if hasattr(self, 'description'):
            stmts.append(self.comment())
        return stmts

    def diff_map(self, infunction):
        """Generate SQL to transform an existing function

        :param infunction: a YAML map defining the new function
        :return: list of SQL statements

        Compares the function to an input function and generates SQL
        statements to transform it into the one represented by the
        input.
        """
        stmts = []
        if self.source != infunction.source:
            stmts.append(self.create(infunction.source))
        stmts.append(self.diff_description(infunction))
        return stmts


class Aggregate(Proc):
    """An aggregate function"""

    objtype = "AGGREGATE"

    def to_map(self):
        """Convert an agggregate to a YAML-suitable format

        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        del dct['language']
        return {self.extern_key(): dct}

    def create(self):
        """Return SQL statements to CREATE the aggregate

        :return: SQL statements
        """
        stmts = []
        ffunc = cond = ''
        if hasattr(self, 'finalfunc'):
            ffname = self.finalfunc[:self.finalfunc.index('(')]
            ffunc = ",\n    FINALFUNC = %s" % (ffname)
        if hasattr(self, 'initcond'):
            cond = ",\n    INITCOND = '%s'" % (self.initcond)
        stmts.append("CREATE AGGREGATE %s(%s) (\n    SFUNC = %s,"
                     "\n    STYPE = %s%s%s)" % (
                self.qualname(),
                self.arguments, self.sfunc, self.stype, ffunc, cond))
        if hasattr(self, 'description'):
            stmts.append(self.comment())
        return stmts


class ProcDict(DbObjectDict):
    "The collection of regular and aggregate functions in a database"

    cls = Proc
    query = \
        """SELECT nspname AS schema, proname AS name,
                  pg_get_function_arguments(p.oid) AS arguments,
                  pg_get_function_result(p.oid) AS returns,
                  l.lanname AS language, provolatile AS volatility,
                  proisstrict AS strict, proisagg, prosrc AS source,
                  prosecdef AS security_definer,
                  aggtransfn::regprocedure AS sfunc,
                  aggtranstype::regtype AS stype,
                  aggfinalfn::regprocedure AS finalfunc,
                  agginitval AS initcond,
                  description
           FROM pg_proc p
                JOIN pg_namespace n ON (pronamespace = n.oid)
                JOIN pg_language l ON (prolang = l.oid)
                LEFT JOIN pg_aggregate a ON (p.oid = aggfnoid)
                LEFT JOIN pg_description d
                     ON (p.oid = d.objoid AND d.objsubid = 0)
           WHERE (nspname != 'pg_catalog' AND nspname != 'information_schema')
           ORDER BY nspname, proname"""

    def _from_catalog(self):
        """Initialize the dictionary of procedures by querying the catalogs"""
        for proc in self.fetch():
            sch, prc, arg = proc.key()
            if hasattr(proc, 'proisagg'):
                del proc.proisagg
                del proc.source
                del proc.volatility
                del proc.returns
                if proc.finalfunc == '-':
                    del proc.finalfunc
                self[(sch, prc, arg)] = Aggregate(**proc.__dict__)
            else:
                self[(sch, prc, arg)] = Function(**proc.__dict__)

    def from_map(self, schema, infuncs):
        """Initalize the dictionary of functions by converting the input map

        :param schema: schema owning the functions
        :param infuncs: YAML map defining the functions
        """
        for key in infuncs.keys():
            spc = key.find(' ')
            if spc == -1:
                raise KeyError("Unrecognized object type: %s" % key)
            objtype = key[:spc]
            if objtype not in ['function', 'aggregate']:
                raise KeyError("Unrecognized object type: %s" % key)
            fnc = key[spc + 1:]
            paren = fnc.find('(')
            if paren == -1 or fnc[-1:] != ')':
                raise KeyError("Invalid function signature: %s" % fnc)
            arguments = fnc[paren + 1:-1]
            infunc = infuncs[key]
            fnc = fnc[:paren]
            if objtype == 'function':
                self[(schema.name, fnc, arguments)] = func = Function(
                    schema=schema.name, name=fnc, arguments=arguments)
            else:
                self[(schema.name, fnc, arguments)] = func = Aggregate(
                    schema=schema.name, name=fnc, arguments=arguments)
                func.language = 'internal'
            if not infunc:
                raise ValueError("Function '%s' has no specification" % fnc)
            for attr, val in infunc.items():
                setattr(func, attr, val)
            if hasattr(func, 'volatility'):
                func.volatility = func.volatility[:1].lower()
            if 'oldname' in infunc:
                func.oldname = infunc['oldname']
            if 'description' in infunc:
                func.description = infunc['description']

    def diff_map(self, infuncs):
        """Generate SQL to transform existing functions

        :param infuncs: a YAML map defining the new functions
        :return: list of SQL statements

        Compares the existing function definitions, as fetched from
        the catalogs, to the input map and generates SQL statements to
        transform the functions accordingly.
        """
        stmts = []
        created = False
        # check input functions
        for (sch, fnc, arg) in infuncs.keys():
            infunc = infuncs[(sch, fnc, arg)]
            if isinstance(infunc, Aggregate):
                continue
            # does it exist in the database?
            if (sch, fnc, arg) not in self:
                if not hasattr(infunc, 'oldname'):
                    # create new function
                    stmts.append(infunc.create())
                    created = True
                else:
                    stmts.append(self[(sch, fnc, arg)].rename(infunc))
            else:
                # check function objects
                diff_stmts = self[(sch, fnc, arg)].diff_map(infunc)
                for stmt in diff_stmts:
                    if isinstance(stmt, list) and stmt:
                        stmt = stmt[0]
                    if isinstance(stmt, basestring) and \
                            stmt.startswith("CREATE "):
                        created = True
                        break
                stmts.append(diff_stmts)

        # check input aggregates
        for (sch, fnc, arg) in infuncs.keys():
            infunc = infuncs[(sch, fnc, arg)]
            if not isinstance(infunc, Aggregate):
                continue
            # does it exist in the database?
            if (sch, fnc, arg) not in self:
                if not hasattr(infunc, 'oldname'):
                    # create new function
                    stmts.append(infunc.create())
                else:
                    stmts.append(self[(sch, fnc, arg)].rename(infunc))
            else:
                # check function objects
                stmts.append(self[(sch, fnc, arg)].diff_map(infunc))

        # check existing functions
        for (sch, fnc, arg) in self.keys():
            func = self[(sch, fnc, arg)]
            # if missing, mark it for dropping
            if (sch, fnc, arg) not in infuncs:
                func.dropped = False

        if created:
            stmts.insert(0, "SET check_function_bodies = false")
        return stmts

    def _drop(self):
        """Actually drop the functions

        :return: SQL statements
        """
        stmts = []
        for (sch, fnc, arg) in self.keys():
            func = self[(sch, fnc, arg)]
            if hasattr(func, 'dropped'):
                stmts.append(func.drop())
        return stmts



# -*- coding: utf-8 -*-
"""Test views"""

import unittest

from utils import PyrseasTestCase, fix_indent, new_std_map

DROP_STMT = "DROP VIEW IF EXISTS v1"


class ViewToMapTestCase(PyrseasTestCase):
    """Test mapping of created views"""

    def test_map_view_no_table(self):
        "Map a created view without a table dependency"
        ddlstmt = "CREATE VIEW v1 AS SELECT now()::date AS today"
        expmap = {'definition': " SELECT now()::date AS today;"}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['view v1'], expmap)

    def test_map_view(self):
        "Map a created view with a table dependency"
        self.db.execute("CREATE TABLE t1 (c1 INTEGER, c2 TEXT, c3 INTEGER)")
        ddlstmt = "CREATE VIEW v1 AS SELECT c1, c3 * 2 FROM t1"
        expmap = {'definition': " SELECT t1.c1, t1.c3 * 2\n   FROM t1;"}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['view v1'], expmap)


class ViewToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input views"""

    def test_create_view_no_table(self):
        "Create a view with no table dependency"
        self.db.execute_commit(DROP_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'view v1': {
                    'definition': " SELECT now()::date AS today;"}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), "CREATE VIEW v1 AS "
                         "SELECT now()::date AS today")

    def test_create_view(self):
        "Create a view"
        self.db.execute_commit(DROP_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}},
                                {'c3': {'type': 'integer'}}]}})
        inmap['schema public'].update({'view v1': {
                    'definition': "SELECT c1, c3 * 2 FROM t1"}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), "CREATE TABLE t1 (c1 integer, "
                         "c2 text, c3 integer)")
        self.assertEqual(fix_indent(dbsql[1]), "CREATE VIEW v1 AS "
                         "SELECT c1, c3 * 2 FROM t1")

    def test_create_view_in_schema(self):
        "Create a view within a non-public schema"
        self.db.execute("CREATE SCHEMA s1")
        self.db.execute_commit(DROP_STMT)
        inmap = new_std_map()
        inmap.update({'schema s1': {'view v1': {
                        'definition': " SELECT now()::date AS today;"}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), "CREATE VIEW s1.v1 AS "
                         "SELECT now()::date AS today")
        self.db.execute_commit("DROP SCHEMA s1 CASCADE")

    def test_bad_view_map(self):
        "Error creating a view with a bad map"
        inmap = new_std_map()
        inmap['schema public'].update({'v1': {
                    'definition': " SELECT now()::date AS today;"}})
        self.assertRaises(KeyError, self.db.process_map, inmap)

    def test_drop_view_no_table(self):
        "Drop an existing view without a table dependency"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE VIEW v1 AS SELECT now()::date AS today")
        dbsql = self.db.process_map(new_std_map())
        self.assertEqual(dbsql, ["DROP VIEW v1"])

    def test_drop_view(self):
        "Drop an existing view with table dependencies"
        self.db.execute(DROP_STMT)
        self.db.execute("CREATE TABLE t1 (c1 INTEGER, c2 TEXT)")
        self.db.execute("CREATE TABLE t2 (c1 INTEGER, c3 TEXT)")
        self.db.execute_commit("CREATE VIEW v1 AS SELECT t1.c1, c2, c3 "
                               "FROM t1 JOIN t2 ON (t1.c1 = t2.c1)")
        dbsql = self.db.process_map(new_std_map())
        self.assertEqual(dbsql[0], "DROP VIEW v1")
        # can't control which table will be dropped first
        drt1 = 1
        drt2 = 2
        if 't1' in dbsql[2]:
            drt1 = 2
            drt2 = 1
        self.assertEqual(dbsql[drt1], "DROP TABLE t1")
        self.assertEqual(dbsql[drt2], "DROP TABLE t2")

    def test_rename_view(self):
        "Rename an existing view"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE VIEW v1 AS SELECT now()::date AS today")
        inmap = new_std_map()
        inmap['schema public'].update({'view v2': {
                    'oldname': 'v1',
                    'definition': " SELECT now()::date AS today;"}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["ALTER VIEW v1 RENAME TO v2"])

    def test_bad_rename_view(self):
        "Error renaming a non-existing view"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE VIEW v1 AS SELECT now()::date AS today")
        inmap = new_std_map()
        inmap['schema public'].update({'view v2': {
                    'oldname': 'v3',
                    'definition': " SELECT now()::date AS today;"}})
        self.assertRaises(KeyError, self.db.process_map, inmap)

    def test_change_view_defn(self):
        "Change view definition"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE VIEW v1 AS SELECT now()::date AS today")
        inmap = new_std_map()
        inmap['schema public'].update({'view v1': {
                    'definition': " SELECT now()::date AS todays_date;"}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), "CREATE OR REPLACE VIEW v1 AS "
                         "SELECT now()::date AS todays_date")


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(ViewToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            ViewToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
    pyrseas.table
    ~~~~~~~~~~~~~

    This module defines three classes: DbType derived from
    DbSchemaObject, Domain derived from DbType, and DbTypeDict derived
    from DbObjectDict.
"""
from pyrseas.dbobject import DbObjectDict, DbSchemaObject
from pyrseas.dbobject.constraint import CheckConstraint


class DbType(DbSchemaObject):
    """A domain or enum type"""

    keylist = ['schema', 'name']


class Enum(DbType):
    "An enumerated type definition"

    objtype = "TYPE"

    def to_map(self):
        """Convert an enum to a YAML-suitable format

        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        return {self.extern_key(): dct}

    def create(self):
        """Return SQL statements to CREATE the enum

        :return: SQL statements
        """
        stmts = []
        lbls = ["'%s'" % lbl for lbl in self.labels]
        stmts.append("CREATE TYPE %s AS ENUM (%s)" % (
                self.qualname(), ",\n    ".join(lbls)))
        if hasattr(self, 'description'):
            stmts.append(self.comment())
        return stmts


class Domain(DbType):
    "A domain definition"

    objtype = "DOMAIN"

    def to_map(self):
        """Convert a domain to a YAML-suitable format

        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        if hasattr(self, 'check_constraints'):
            if not 'check_constraints' in dct:
                dct.update(check_constraints={})
            for cns in self.check_constraints.values():
                dct['check_constraints'].update(
                    self.check_constraints[cns.name].to_map(None))

        return {self.extern_key(): dct}

    def create(self):
        """Return SQL statements to CREATE the domain

        :return: SQL statements
        """
        stmts = []
        create = "CREATE DOMAIN %s AS %s" % (self.qualname(), self.type)
        if hasattr(self, 'not_null'):
            create += ' NOT NULL'
        if hasattr(self, 'default'):
            create += ' DEFAULT ' + str(self.default)
        if hasattr(self, 'check_constraints'):
            cnslist = []
            for cns in self.check_constraints.values():
                cnslist.append(" CONSTRAINT %s CHECK (%s)" % (
                    cns.name, cns.expression))
            create += ", ".join(cnslist)
        stmts.append(create)
        if hasattr(self, 'description'):
            stmts.append(self.comment())
        return stmts

    def diff_map(self, indomain):
        """Generate SQL to transform an existing domain

        :param indomain: a YAML map defining the new domain
        :return: list of SQL statements

        Compares the domain to an input domain and generates SQL
        statements to transform it into the one represented by the
        input.
        """
        stmts = []
        stmts.append(self.diff_description(indomain))
        return stmts


class TypeDict(DbObjectDict):
    "The collection of domains and enums in a database"

    cls = DbType
    query = \
        """SELECT nspname AS schema, typname AS name, typtype AS kind,
                  format_type(typbasetype, typtypmod) AS type,
                  typnotnull AS not_null, typdefault AS default,
                  ARRAY(SELECT enumlabel FROM pg_enum e WHERE t.oid = enumtypid
                  ORDER BY e.oid) AS labels,
                  description
           FROM pg_type t
                JOIN pg_namespace n ON (typnamespace = n.oid)
                LEFT JOIN pg_description d
                     ON (t.oid = d.objoid AND d.objsubid = 0)
           WHERE typtype in ('d', 'e')
             AND (nspname != 'pg_catalog' AND nspname != 'information_schema')
           ORDER BY nspname, typname"""

    def _from_catalog(self):
        """Initialize the dictionary of types by querying the catalogs"""
        for dbtype in self.fetch():
            sch, typ = dbtype.key()
            kind = dbtype.kind
            del dbtype.kind
            if kind == 'd':
                self[(sch, typ)] = Domain(**dbtype.__dict__)
            elif kind == 'e':
                del dbtype.type
                self[(sch, typ)] = Enum(**dbtype.__dict__)

    def from_map(self, schema, inobjs, newdb):
        """Initalize the dictionary of types by converting the input map

        :param schema: schema owning the types
        :param inobjs: YAML map defining the schema objects
        :param newdb: collection of dictionaries defining the database
        """
        for k in inobjs.keys():
            spc = k.find(' ')
            if spc == -1:
                raise KeyError("Unrecognized object type: %s" % k)
            objtype = k[:spc]
            key = k[spc + 1:]
            if objtype == 'domain':
                self[(schema.name, key)] = domain = Domain(
                    schema=schema.name, name=key)
                indomain = inobjs[k]
                if not indomain:
                    raise ValueError("Domain '%s' has no specification" % k)
                for attr, val in indomain.items():
                    setattr(domain, attr, val)
                if 'oldname' in indomain:
                    domain.oldname = indomain['oldname']
                newdb.constraints.from_map(domain, indomain, 'd')
                if 'description' in indomain:
                    domain.description = indomain['description']
            elif objtype == 'type':
                if 'labels' in inobjs[k]:
                    self[(schema.name, key)] = enum = Enum(
                        schema=schema.name, name=key)
                    inenum = inobjs[k]
                    enum.labels = inenum['labels']
                    if 'oldname' in inenum:
                        enum.oldname = inenum['oldname']
                    if 'description' in inenum:
                        enum.description = inenum['description']
            else:
                raise KeyError("Unrecognized object type: %s" % k)

    def link_refs(self, dbconstrs):
        """Connect constraints to their respective domains

        :param dbconstrs: dictionary of constraints

        Fills the `check_constraints` dictionaries for each domain by
        traversing the `dbconstrs` dictionary.
        """
        for (sch, typ, cns) in dbconstrs.keys():
            constr = dbconstrs[(sch, typ, cns)]
            if not hasattr(constr, 'target') or constr.target != 'd':
                continue
            assert self[(sch, typ)]
            dbtype = self[(sch, typ)]
            if isinstance(constr, CheckConstraint):
                if not hasattr(dbtype, 'check_constraints'):
                    dbtype.check_constraints = {}
                dbtype.check_constraints.update({cns: constr})

    def diff_map(self, intypes):
        """Generate SQL to transform existing domains and types

        :param intypes: a YAML map defining the new domains/types
        :return: list of SQL statements

        Compares the existing domain/type definitions, as fetched from
        the catalogs, to the input map and generates SQL statements to
        transform the domains/types accordingly.
        """
        stmts = []
        # check input types
        for (sch, typ) in intypes.keys():
            intype = intypes[(sch, typ)]
            # does it exist in the database?
            if (sch, typ) not in self:
                if not hasattr(intype, 'oldname'):
                    # create new type
                    stmts.append(intype.create())
                else:
                    stmts.append(self[(sch, intype.oldname)].rename(typ))
                    del self[(sch, intype.oldname)]

        # check existing types
        for (sch, typ) in self.keys():
            dbtype = self[(sch, typ)]
            # if missing, mark it for dropping
            if (sch, typ) not in intypes:
                dbtype.dropped = False
            else:
                # check type/sequence/view objects
                stmts.append(dbtype.diff_map(intypes[(sch, typ)]))

        return stmts

    def _drop(self):
        """Actually drop the types

        :return: SQL statements
        """
        stmts = []
        for (sch, typ) in self.keys():
            dbtype = self[(sch, typ)]
            if hasattr(dbtype, 'dropped'):
                stmts.append(dbtype.drop())
        return stmts



# -*- coding: utf-8 -*-
"""Test enums and other types"""

import unittest

from utils import PyrseasTestCase, fix_indent, new_std_map

CREATE_ENUM_STMT = "CREATE TYPE t1 AS ENUM ('red', 'green', 'blue')"
DROP_STMT = "DROP TYPE IF EXISTS t1"
COMMENT_STMT = "COMMENT ON TYPE t1 IS 'Test type t1'"


class EnumToMapTestCase(PyrseasTestCase):
    """Test mapping of created enum types"""

    def test_enum(self):
        "Map an enum"
        self.db.execute_commit(DROP_STMT)
        expmap = {'labels': ['red', 'green', 'blue']}
        dbmap = self.db.execute_and_map(CREATE_ENUM_STMT)
        self.assertEqual(dbmap['schema public']['type t1'], expmap)


class EnumToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input enums"""

    def test_create_enum(self):
        "Create an enum"
        self.db.execute_commit(DROP_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'type t1': {
                    'labels': ['red', 'green', 'blue']}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), CREATE_ENUM_STMT)

    def test_drop_enum(self):
        "Drop an existing enum"
        self.db.execute(DROP_STMT)
        self.db.execute_commit(CREATE_ENUM_STMT)
        inmap = new_std_map()
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["DROP TYPE t1"])

    def test_rename_enum(self):
        "Rename an existing enum"
        self.db.execute(DROP_STMT)
        self.db.execute_commit(CREATE_ENUM_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({'type t2': {
                    'oldname': 't1', 'labels': ['red', 'green', 'blue']}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["ALTER TYPE t1 RENAME TO t2"])


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(EnumToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            EnumToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""Test languages"""

import unittest

from utils import PyrseasTestCase

CREATE_STMT = "CREATE LANGUAGE plperl"
DROP_STMT = "DROP LANGUAGE IF EXISTS plperl CASCADE"


class LanguageToMapTestCase(PyrseasTestCase):
    """Test mapping of existing languages"""

    def test_map_language(self):
        "Map an existing language"
        self.db.execute(DROP_STMT)
        dbmap = self.db.execute_and_map(CREATE_STMT)
        self.assertEqual(dbmap['language plperl'], {'trusted': True})


class LanguageToSqlTestCase(PyrseasTestCase):
    """Test SQL generation for input languages"""

    def tearDown(self):
        self.db.execute_commit(DROP_STMT)
        self.db.close()

    def test_create_language(self):
        "Create a language that didn't exist"
        self.db.execute_commit(DROP_STMT)
        dbsql = self.db.process_map({'language plperl': {}})
        self.assertEqual(dbsql, [CREATE_STMT])

    def test_bad_language_map(self):
        "Error creating a language with a bad map"
        self.assertRaises(KeyError, self.db.process_map, {'plperl': {}})

    def test_drop_language(self):
        "Drop an existing language"
        self.db.execute(DROP_STMT)
        self.db.execute_commit(CREATE_STMT)
        dbsql = self.db.process_map({})
        self.assertEqual(dbsql, ["DROP LANGUAGE plperl"])

    def test_drop_language_function(self):
        "Drop an existing function and the language it uses"
        self.db.execute(DROP_STMT)
        self.db.execute(CREATE_STMT)
        self.db.execute_commit("CREATE FUNCTION f1() RETURNS text "
                               "LANGUAGE plperl AS $_$return \"dummy\";$_$")
        dbsql = self.db.process_map({})
        self.assertEqual(dbsql, ["DROP FUNCTION f1()", "DROP LANGUAGE plperl"])


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(LanguageToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            LanguageToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""Test functions"""

import unittest

from utils import PyrseasTestCase, fix_indent, new_std_map

SOURCE1 = "SELECT 'dummy'::text"
CREATE_STMT1 = "CREATE FUNCTION f1() RETURNS text LANGUAGE sql AS " \
    "$_$%s$_$ IMMUTABLE" % SOURCE1
DROP_STMT1 = "DROP FUNCTION IF EXISTS f1()"
SOURCE2 = "SELECT GREATEST($1, $2)"
CREATE_STMT2 = "CREATE FUNCTION f1(integer, integer) RETURNS integer " \
    "LANGUAGE sql AS $_$%s$_$" % SOURCE2
DROP_STMT2 = "DROP FUNCTION IF EXISTS f1(integer, integer)"
COMMENT_STMT = "COMMENT ON FUNCTION f1(integer, integer) IS 'Test function f1'"


class FunctionToMapTestCase(PyrseasTestCase):
    """Test mapping of existing functions"""

    def test_map_function(self):
        "Map a very simple function with no arguments"
        expmap = {'language': 'sql', 'returns': 'text',
                  'source': SOURCE1, 'volatility': 'immutable'}
        dbmap = self.db.execute_and_map(CREATE_STMT1)
        self.assertEqual(dbmap['schema public']['function f1()'], expmap)

    def test_map_function_with_args(self):
        "Map a function with two arguments"
        expmap = {'language': 'sql', 'returns': 'integer', 'source': SOURCE2}
        dbmap = self.db.execute_and_map(CREATE_STMT2)
        self.assertEqual(dbmap['schema public'] \
                             ['function f1(integer, integer)'], expmap)

    def test_map_void_function(self):
        "Map a function returning void"
        self.db.execute("CREATE TABLE t1 (c1 integer, c2 text)")
        expmap = {'language': 'sql', 'returns': 'void',
                  'source': "INSERT INTO t1 VALUES (1, 'dummy')"}
        dbmap = self.db.execute_and_map(
            "CREATE FUNCTION f1() RETURNS void LANGUAGE sql AS "
            "$_$INSERT INTO t1 VALUES (1, 'dummy')$_$")
        self.assertEqual(dbmap['schema public']['function f1()'], expmap)

    def test_map_setof_row_function(self):
        "Map a function returning a set of rows"
        self.db.execute("CREATE TABLE t1 (c1 integer, c2 text)")
        expmap = {'language': 'sql', 'returns': 'SETOF t1',
                  'source': "SELECT * FROM t1"}
        dbmap = self.db.execute_and_map(
            "CREATE FUNCTION f1() RETURNS SETOF t1 LANGUAGE sql AS "
            "$_$SELECT * FROM t1$_$")
        self.assertEqual(dbmap['schema public']['function f1()'], expmap)

    def test_map_security_definer_function(self):
        "Map a function that is SECURITY DEFINER"
        expmap = {'language': 'sql', 'returns': 'text',
                  'source': SOURCE1, 'security_definer': True}
        dbmap = self.db.execute_and_map("CREATE FUNCTION f1() RETURNS text "
                                        "LANGUAGE sql SECURITY DEFINER AS "
                                        "$_$%s$_$" % SOURCE1)
        self.assertEqual(dbmap['schema public']['function f1()'], expmap)

    def test_map_function_comment(self):
        "Map a function comment"
        self.db.execute(CREATE_STMT2)
        dbmap = self.db.execute_and_map(COMMENT_STMT)
        self.assertEqual(dbmap['schema public'] \
                             ['function f1(integer, integer)']['description'],
                         'Test function f1')


class FunctionToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input functions"""

    def test_create_function(self):
        "Create a very simple function with no arguments"
        self.db.execute_commit(DROP_STMT1)
        inmap = new_std_map()
        inmap['schema public'].update({'function f1()': {
                    'language': 'sql', 'returns': 'text',
                    'source': SOURCE1, 'volatility': 'immutable'}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[1]), CREATE_STMT1)

    def test_create_function_with_args(self):
        "Create a function with two arguments"
        self.db.execute_commit(DROP_STMT2)
        inmap = new_std_map()
        inmap['schema public'].update({
                'function f1(integer, integer)': {
                    'language': 'sql', 'returns': 'integer',
                    'source': SOURCE2}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[1]), CREATE_STMT2)

    def test_create_setof_row_function(self):
        "Create a function returning a set of rows"
        self.db.execute_commit(DROP_STMT1)
        inmap = new_std_map()
        inmap['schema public'].update({'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}]}})
        inmap['schema public'].update({
                'function f1()': {
                    'language': 'sql', 'returns': 'SETOF t1',
                    'source': "SELECT * FROM t1"}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[2]),
                         "CREATE FUNCTION f1() RETURNS SETOF t1 LANGUAGE sql "
                         "AS $_$SELECT * FROM t1$_$")

    def test_create_security_definer_function(self):
        "Create a SECURITY DEFINER function"
        self.db.execute_commit(DROP_STMT1)
        inmap = new_std_map()
        inmap['schema public'].update({'function f1()': {
                    'language': 'sql', 'returns': 'text',
                    'source': SOURCE1, 'security_definer': True}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[1]),
                         "CREATE FUNCTION f1() RETURNS text LANGUAGE sql "
                         "AS $_$%s$_$ SECURITY DEFINER" % SOURCE1)

    def test_create_function_in_schema(self):
        "Create a function within a non-public schema"
        self.db.execute("CREATE SCHEMA s1")
        self.db.execute_commit(DROP_STMT1)
        inmap = new_std_map()
        inmap.update({'schema s1': {'function f1()': {
                    'language': 'sql', 'returns': 'text',
                    'source': SOURCE1, 'volatility': 'immutable'}}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[1]), "CREATE FUNCTION s1.f1() "
                         "RETURNS text LANGUAGE sql "
                         "AS $_$%s$_$ IMMUTABLE" % SOURCE1)
        self.db.execute_commit("DROP SCHEMA s1 CASCADE")

    def test_bad_function_map(self):
        "Error creating a function with a bad map"
        inmap = new_std_map()
        inmap['schema public'].update({'f1()': {
                    'language': 'sql', 'returns': 'text',
                    'source': SOURCE1}})
        self.assertRaises(KeyError, self.db.process_map, inmap)

    def test_drop_function(self):
        "Drop an existing function with no arguments"
        self.db.execute(DROP_STMT1)
        self.db.execute_commit(CREATE_STMT1)
        dbsql = self.db.process_map(new_std_map())
        self.assertEqual(dbsql, ["DROP FUNCTION f1()"])

    def test_drop_function_with_args(self):
        "Drop an existing function which has arguments"
        self.db.execute(DROP_STMT2)
        self.db.execute_commit(CREATE_STMT2)
        dbsql = self.db.process_map(new_std_map())
        self.assertEqual(dbsql[0], "DROP FUNCTION f1(integer, integer)")

    def test_change_function_defn(self):
        "Change function definition"
        self.db.execute(DROP_STMT1)
        self.db.execute_commit(CREATE_STMT1)
        inmap = new_std_map()
        inmap['schema public'].update({'function f1()': {
                    'language': 'sql', 'returns': 'text',
                    'source': "SELECT 'example'::text",
                    'volatility': 'immutable'}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[1]), "CREATE OR REPLACE "
                         "FUNCTION f1() RETURNS text LANGUAGE sql "
                         "AS $_$SELECT 'example'::text$_$ IMMUTABLE")

    def test_function_with_comment(self):
        "Create a function with a comment"
        self.db.execute_commit(DROP_STMT2)
        inmap = new_std_map()
        inmap['schema public'].update({
                'function f1(integer, integer)': {
                    'description': 'Test function f1',
                    'language': 'sql', 'returns': 'integer',
                    'source': SOURCE2}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[1]), CREATE_STMT2)
        self.assertEqual(dbsql[2], COMMENT_STMT)

    def test_comment_on_function(self):
        "Create a comment for an existing function"
        self.db.execute(DROP_STMT2)
        self.db.execute_commit(CREATE_STMT2)
        inmap = new_std_map()
        inmap['schema public'].update({
                'function f1(integer, integer)': {
                    'description': 'Test function f1',
                    'language': 'sql', 'returns': 'integer',
                    'source': SOURCE2}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, [COMMENT_STMT])

    def test_drop_function_comment(self):
        "Drop a comment on an existing function"
        self.db.execute(DROP_STMT2)
        self.db.execute(CREATE_STMT2)
        self.db.execute_commit(COMMENT_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({
                'function f1(integer, integer)': {
                    'language': 'sql', 'returns': 'integer',
                    'source': SOURCE2}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql,
                         ["COMMENT ON FUNCTION f1(integer, integer) IS NULL"])

    def test_change_function_comment(self):
        "Change existing comment on a function"
        self.db.execute(DROP_STMT2)
        self.db.execute(CREATE_STMT2)
        self.db.execute_commit(COMMENT_STMT)
        inmap = new_std_map()
        inmap['schema public'].update({
                'function f1(integer, integer)': {
                    'description': 'Changed function f1',
                    'language': 'sql', 'returns': 'integer',
                    'source': SOURCE2}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, [
                "COMMENT ON FUNCTION f1(integer, integer) IS "
                "'Changed function f1'"])


class AggregateToMapTestCase(PyrseasTestCase):
    """Test mapping of existing aggregates"""

    def test_map_aggregate(self):
        "Map a simple aggregate"
        self.db.execute(CREATE_STMT2 + " IMMUTABLE")
        expmap = {'sfunc': 'f1(integer,integer)', 'stype': 'integer'}
        dbmap = self.db.execute_and_map("CREATE AGGREGATE a1 (integer) ("
                                        "SFUNC = f1, STYPE = integer)")
        self.assertEqual(dbmap['schema public'][
                'function f1(integer, integer)'], {
                'language': 'sql', 'returns': 'integer',
                'source': SOURCE2,
                'volatility': 'immutable'})
        self.assertEqual(dbmap['schema public'][
                'aggregate a1(integer)'], expmap)

    def test_map_aggregate_init_final(self):
        "Map an aggregate with an INITCOND and a FINALFUNC"
        self.db.execute(CREATE_STMT2 + " IMMUTABLE")
        self.db.execute("CREATE FUNCTION f2(integer) RETURNS float "
                        "LANGUAGE sql AS $_$SELECT $1::float$_$ IMMUTABLE")
        expmap = {'sfunc': 'f1(integer,integer)', 'stype': 'integer',
                  'initcond': '-1', 'finalfunc': 'f2(integer)'}
        dbmap = self.db.execute_and_map(
            "CREATE AGGREGATE a1 (integer) (SFUNC = f1, STYPE = integer, "
            "FINALFUNC = f2, INITCOND = '-1')")
        self.assertEqual(dbmap['schema public'][
                'function f1(integer, integer)'], {
                'language': 'sql', 'returns': 'integer',
                'source': SOURCE2,
                'volatility': 'immutable'})
        self.assertEqual(dbmap['schema public']['function f2(integer)'],
                         {'language': 'sql', 'returns': 'double precision',
                          'source': "SELECT $1::float",
                          'volatility': 'immutable'})
        self.assertEqual(dbmap['schema public']['aggregate a1(integer)'],
                         expmap)


class AggregateToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input aggregates"""

    def test_create_aggregate(self):
        "Create a simple aggregate"
        self.db.execute("DROP AGGREGATE IF EXISTS a1(integer)")
        self.db.execute_commit(DROP_STMT2)
        inmap = new_std_map()
        inmap['schema public'].update({'function f1(integer, integer)': {
                    'language': 'sql', 'returns': 'integer',
                    'source': SOURCE2,
                    'volatility': 'immutable'}})
        inmap['schema public'].update({'aggregate a1(integer)': {
                    'sfunc': 'f1', 'stype': 'integer'}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[1]), CREATE_STMT2 + " IMMUTABLE")
        self.assertEqual(fix_indent(dbsql[2]),
                         "CREATE AGGREGATE a1(integer) "
                         "(SFUNC = f1, STYPE = integer)")

    def test_create_aggregate_init_final(self):
        "Create an aggregate with an INITCOND and a FINALFUNC"
        self.db.execute("DROP AGGREGATE IF EXISTS a1(integer)")
        self.db.execute_commit("DROP FUNCTION IF EXISTS f2(integer)")
        self.db.execute_commit(DROP_STMT2)
        inmap = new_std_map()
        inmap['schema public'].update({'function f1(integer, integer)': {
                    'language': 'sql', 'returns': 'integer',
                    'source': SOURCE2,
                    'volatility': 'immutable'}})
        inmap['schema public'].update({'function f2(integer)': {
                    'language': 'sql', 'returns': 'double precision',
                    'source': "SELECT $1::float",
                    'volatility': 'immutable'}})
        inmap['schema public'].update({'aggregate a1(integer)': {
                    'sfunc': 'f1', 'stype': 'integer', 'initcond': '-1',
                    'finalfunc': 'f2(integer)'}})
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[1]), CREATE_STMT2 + " IMMUTABLE")
        self.assertEqual(fix_indent(dbsql[2]),
                         "CREATE FUNCTION f2(integer) "
                         "RETURNS double precision LANGUAGE sql "
                         "AS $_$SELECT $1::float$_$ IMMUTABLE")
        self.assertEqual(fix_indent(dbsql[3]),
                         "CREATE AGGREGATE a1(integer) "
                         "(SFUNC = f1, STYPE = integer, FINALFUNC = f2, "
                         "INITCOND = '-1')")


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(FunctionToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            FunctionToSqlTestCase))
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            AggregateToMapTestCase))
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            AggregateToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')
